import time

from sc3nb.sc_objects.node import Synth
from tests.conftest import SCBaseTest


class SynthTestWithSClang(SCBaseTest):
    __test__ = True
    start_sclang = True

    def setUp(self) -> None:
        self.assertIsNotNone(SynthTestWithSClang.sc.lang)
        self.custom_nodeid = 42
        self.all_synth_args = {"freq": 400, "amp": 0.3, "num": 4, "pan": 0, "lg": 0.1}
        self.synth = Synth("s2", nodeid=self.custom_nodeid)

    def tearDown(self) -> None:
        self.synth.free()
        time.sleep(0.1)

    def test_synth_desc(self):
        self.assertIsNotNone(self.synth.synth_desc)
        for name, synth_arg in self.synth.synth_desc.items():
            self.assertEqual(name, synth_arg.name)
            self.assertAlmostEqual(self.synth.get(name), synth_arg.default)

    def test_getattr(self):
        for name, value in self.all_synth_args.items():
            self.assertAlmostEqual(self.synth.__getattr__(name), value)



"""Contains classes enabling the easy communication with SuperCollider
within jupyter notebooks
"""

import logging
from typing import Optional, Sequence

from IPython import get_ipython

import sc3nb.magics as magics
from sc3nb.process_handling import ALLOWED_PARENTS
from sc3nb.sc_objects.server import SCServer, ServerOptions
from sc3nb.sclang import SCLang

_LOGGER = logging.getLogger(__name__)


def startup(
    start_server: bool = True,
    scsynth_path: Optional[str] = None,
    start_sclang: bool = True,
    sclang_path: Optional[str] = None,
    magic: bool = True,
    scsynth_options: Optional[ServerOptions] = None,
    with_blip: bool = True,
    console_logging: bool = False,
    allowed_parents: Sequence[str] = ALLOWED_PARENTS,
    timeout: float = 10,
) -> "SC":
    """Inits SuperCollider (scsynth, sclang) and registers ipython magics

    Parameters
    ----------
    start_server : bool, optional
        If True boot scsynth, by default True
    scsynth_path : Optional[str], optional
        Path of scscynth executable, by default None
    start_sclang : bool, optional
        If True start sclang, by default True
    sclang_path : Optional[str], optional
        Path of sclang executable, by default None
    magic : bool, optional
        If True register magics to ipython, by default True
    scsynth_options : Optional[ServerOptions], optional
        Options for the server, by default None
    with_blip : bool, optional
            make a sound when booted, by default True
    console_logging : bool, optional
        If True write scsynth/sclang output to console, by default False
    allowed_parents : Sequence[str], optional
        Names of parents that are allowed for other instances of
        sclang/scsynth processes, by default ALLOWED_PARENTS
    timeout : float, optional
        timeout in seconds for starting the executable, by default 10

    Returns
    -------
    SC
        SuperCollider Interface class.
    """
    if magic:
        ipy = get_ipython()
        if ipy is not None:
            magics.load_ipython_extension(ipy)

    if SC.default is None:
        SC.default = SC(
            start_server=start_server,
            scsynth_path=scsynth_path,
            start_sclang=start_sclang,
            sclang_path=sclang_path,
            scsynth_options=scsynth_options,
            with_blip=with_blip,
            console_logging=console_logging,
            allowed_parents=allowed_parents,
            timeout=timeout,
        )
    else:
        _LOGGER.warning("SC already started")
        if start_sclang:
            SC.get_default().start_sclang(
                sclang_path=sclang_path,
                allowed_parents=allowed_parents,
                timeout=timeout,
            )
        if start_server:
            SC.get_default().start_server(
                scsynth_options=scsynth_options,
                scsynth_path=scsynth_path,
                allowed_parents=allowed_parents,
                timeout=timeout,
            )
    return SC.default


class SC:
    """Create a SuperCollider Wrapper object.

    Parameters
    ----------
    start_server : bool, optional
        If True boot scsynth, by default True.
    scsynth_path : Optional[str], optional
        Path of scscynth executable, by default None.
    start_sclang : bool, optional
        If True start sclang, by default True.
    sclang_path : Optional[str], optional
        Path of sclang executable, by default None.
    scsynth_options : Optional[ServerOptions], optional
        Options for the server, by default None.
    with_blip : bool, optional
        Make a sound when booted, by default True.
    console_logging : bool, optional
        If True write scsynth/sclang output to console, by default True.
    allowed_parents : Sequence[str], optional
        Names of parents that are allowed for other instances of
        sclang/scsynth processes, by default ALLOWED_PARENTS.
    timeout : float, optional
        timeout in seconds for starting the executables, by default 5
    """

    default: Optional["SC"] = None
    """Default SC instance.

    This will be used by all SuperCollider objects if no SC/server/lang is specified.
    """

    @classmethod
    def get_default(cls) -> "SC":
        """Get the default SC instance

        Returns
        -------
        SC
            default SC instance

        Raises
        ------
        RuntimeError
            If there is no default SC instance.
        """
        if cls.default is not None:
            return cls.default
        else:
            raise RuntimeError(
                "You need to start a SuperCollider SC instance first"
                " or provide an sclang/server directly."
            )

    def __init__(
        self,
        *,
        start_server: bool = True,
        scsynth_path: Optional[str] = None,
        start_sclang: bool = True,
        sclang_path: Optional[str] = None,
        scsynth_options: Optional[ServerOptions] = None,
        with_blip: bool = True,
        console_logging: bool = True,
        allowed_parents: Sequence[str] = ALLOWED_PARENTS,
        timeout: float = 5,
    ):
        if SC.default is None:
            SC.default = self
        self._console_logging = console_logging
        self._server = None
        self._sclang = None
        try:
            if start_sclang:
                self.start_sclang(
                    sclang_path=sclang_path,
                    console_logging=self._console_logging,
                    allowed_parents=allowed_parents,
                    timeout=timeout,
                )
            if start_server:
                self.start_server(
                    scsynth_path=scsynth_path,
                    scsynth_options=scsynth_options,
                    console_logging=self._console_logging,
                    with_blip=with_blip,
                    allowed_parents=allowed_parents,
                    timeout=timeout,
                )
        except Exception:
            if SC.default is self:
                SC.default = None
            raise

    def start_sclang(
        self,
        sclang_path: Optional[str] = None,
        console_logging: bool = True,
        allowed_parents: Sequence[str] = ALLOWED_PARENTS,
        timeout: float = 5,
    ):
        """Start this SuperCollider language

        Parameters
        ----------
        sclang_path : Optional[str], optional
            Path of sclang executable, by default None
        console_logging : bool, optional
            If True write scsynth/sclang output to console, by default True
        allowed_parents : Sequence[str], optional
            Names of parents that are allowed for other instances of
            sclang/scsynth processes, by default ALLOWED_PARENTS
        timeout : float, optional
            timeout in seconds for starting the executable, by default 5
        """
        if self._sclang is None:
            self._sclang = SCLang()
            try:
                self._sclang.start(
                    sclang_path=sclang_path,
                    console_logging=console_logging,
                    allowed_parents=allowed_parents,
                    timeout=timeout,
                )
            except Exception as excep:
                self._sclang = None
                raise RuntimeError(f"Starting sclang failed - {excep}") from excep
            else:
                self._try_to_connect()
        else:
            _LOGGER.warning("sclang already started")
        if SC.default is None:
            SC.default = self

    def start_server(
        self,
        scsynth_options: Optional[ServerOptions] = None,
        scsynth_path: Optional[str] = None,
        console_logging: bool = True,
        with_blip: bool = True,
        allowed_parents: Sequence[str] = ALLOWED_PARENTS,
        timeout: float = 5,
    ):
        """Start this SuperCollider server

        Parameters
        ----------
        scsynth_options : Optional[ServerOptions], optional
            Options for the server, by default None
        scsynth_path : Optional[str], optional
            Path of scscynth executable, by default None
        console_logging : bool, optional
            If True write scsynth/sclang output to console, by default True
        with_blip : bool, optional
            make a sound when booted, by default True
        allowed_parents : Sequence[str], optional
            Names of parents that are allowed for other instances of
            sclang/scsynth processes, by default ALLOWED_PARENTS
        timeout : float, optional
            timeout in seconds for starting the executable, by default 5
        """
        if self._server is None:
            self._server = SCServer(options=scsynth_options)
            try:
                self._server.boot(
                    scsynth_path=scsynth_path,
                    console_logging=console_logging,
                    allowed_parents=allowed_parents,
                    with_blip=with_blip,
                    timeout=timeout,
                )
            except Exception as excep:
                self._server = None
                raise RuntimeError(f"Starting scsynth failed - {excep}") from excep
            else:
                self._try_to_connect()
        else:
            _LOGGER.warning("scsynth already started")
        if SC.default is None:
            SC.default = self

    def _try_to_connect(self):
        if self._sclang is not None and self._server is not None:
            try:
                self._sclang.connect_to_server(self._server)
            except Exception as excep:
                raise RuntimeError(
                    f"connecting {self._sclang} to {self._server} failed"
                ) from excep

    @property
    def server(self) -> SCServer:
        """This SuperCollider server object"""
        if self._server is not None:
            return self._server
        else:
            raise RuntimeError("You need to start the SuperCollider Server first")

    @property
    def lang(self) -> SCLang:
        """This SuperCollider language object"""
        if self._sclang is not None:
            return self._sclang
        else:
            raise RuntimeError(
                "You need to start the SuperCollider Language (sclang) first"
            )

    @property
    def console_logging(self):
        """True if console logging enabled"""
        return self._console_logging

    @console_logging.setter
    def console_logging(self, value):
        if self._server is not None:
            self._server.process.console_logging = value
        if self._sclang is not None:
            self._sclang.process.console_logging = value

    def __del__(self):
        self.exit()

    def __repr__(self) -> str:
        server = self._server if self._server is not None else "SCServer not started"
        lang = self._sclang if self._sclang is not None else "SCLang not started"
        return f"<SC {server},\n    {lang}>"

    def exit(self) -> None:
        """Closes SuperCollider and shuts down server"""
        try:
            if self._server is not None:
                self._server.quit()
            if self._sclang is not None:
                self._sclang.kill()
        finally:
            self._server = None
            self._sclang = None
            if self is SC.default:
                SC.default = None



"""Package for interfacing SuperCollider.

Collection of Classes and functions for communicating
with SuperCollider within python and jupyter notebooks,
as well as playing recording and visualizing audio.

Examples
--------
For example usage please refer to the user guide.
"""


from sc3nb.sc import startup, SC

from sc3nb.sc_objects.server import SCServer, ServerOptions
from sc3nb.sclang import SCLang

from sc3nb.sc_objects.node import Node, Synth, Group, AddAction
from sc3nb.sc_objects.synthdef import SynthDef
from sc3nb.sc_objects.buffer import Buffer
from sc3nb.sc_objects.bus import Bus
from sc3nb.sc_objects.score import Score

from sc3nb.sc_objects.recorder import Recorder

from sc3nb.timed_queue import TimedQueue, TimedQueueSC
from sc3nb.osc.osc_communication import OSCMessage, Bundler

from pyamapping import (
    linlin,
    midicps,
    midi_to_cps,
    cps_to_midi,
    cpsmidi,
    clip,
    db_to_amp,
    dbamp,
    amp_to_db,
    ampdb,
)


__all__ = [
    "startup",
    "SC",
    "TimedQueue",
    "TimedQueueSC",
    "OSCMessage",
    "Bundler",
    "SCLang",
    "linlin",
    "midi_to_cps",
    "midicps",
    "cps_to_midi",
    "cpsmidi",
    "clip",
    "db_to_amp",
    "dbamp",
    "amp_to_db",
    "ampdb",
    "Buffer",
    "SynthDef",
    "Node",
    "Synth",
    "Group",
    "AddAction",
    "SCServer",
    "ServerOptions",
    "Recorder",
    "Score",
    "Bus",
]


def load_ipython_extension(ipython):
    """Load the extension in IPython."""
    from sc3nb.magics import load_ipython_extension as load_extension

    load_extension(ipython)



"""Classes to run register functions at certain timepoints and run asynchronously"""

import threading
import time
from typing import Any, Callable, Iterable, NoReturn, Union

import numpy as np

import sc3nb
from sc3nb.osc.osc_communication import Bundler, OSCCommunication, OSCMessage


class Event:
    """Stores a timestamp, function and arguments for that function.
    Long running functions can be wrapped inside an own thread

    Parameters
    ----------
    timestamp : float
        Time event should be executed
    function : Callable[..., None]
        Function to be executed
    args : Iterable[Any]
        Arguments for function
    spawn : bool, optional
        if True, create new thread for function, by default False
    """

    def __init__(
        self,
        timestamp: float,
        function: Callable[..., None],
        args: Iterable[Any],
        spawn: bool = False,
    ) -> None:
        if spawn:
            thread = threading.Thread(target=function, args=args)
            function = thread.start
            args = ()
        self.timestamp = timestamp
        self.function = function
        self.args = args

    def execute(self) -> None:
        """Executes function"""
        self.function(*self.args)

    def __eq__(self, other):
        return self.timestamp == other.timestamp

    def __lt__(self, other):
        return self.timestamp < other.timestamp

    def __le__(self, other):
        return self.timestamp <= other.timestamp

    def __repr__(self):
        return "%s: %s" % (self.timestamp, self.function.__name__)


class TimedQueue:
    """Accumulates events as timestamps and functions.

    Executes given functions according to the timestamps

    Parameters
    ----------
    relative_time : bool, optional
        If True, use relative time, by default False
    thread_sleep_time : float, optional
        Sleep time in seconds for worker thread, by default 0.001
    drop_time_threshold : float, optional
        Threshold for execution time of events in seconds.
        If this is exceeded the event will be dropped, by default 0.5
    """

    def __init__(
        self,
        relative_time: bool = False,
        thread_sleep_time: float = 0.001,
        drop_time_threshold: float = 0.5,
    ) -> None:
        self.drop_time_thr = drop_time_threshold
        self.start = time.time() if relative_time else 0
        self.onset_idx = np.empty((0, 2))
        self.event_list = []
        self.close_event = threading.Event()

        self.lock = threading.Lock()

        self.thread = threading.Thread(
            target=self.__worker, args=(thread_sleep_time, self.close_event)
        )  # , daemon=True)

        self.thread.start()

    def close(self) -> None:
        """Closes event processing without waiting for pending events"""
        self.close_event.set()
        self.thread.join()

    def join(self) -> None:
        """Closes event processing after waiting for pending events"""
        self.complete()
        self.close_event.set()
        self.thread.join()

    def complete(self) -> None:
        """Blocks until all pending events have completed"""
        while self.event_list:
            time.sleep(0.01)

    def put(
        self,
        timestamp: float,
        function: Callable[..., None],
        args: Iterable[Any] = (),
        spawn: bool = False,
    ) -> None:
        """Adds event to queue

        Parameters
        ----------
        timestamp : float
            Time (POSIX) when event should be executed
        function : Callable[..., None]
            Function to be executed
        args : Iterable[Any], optional
            Arguments to be passed to function, by default ()
        spawn : bool, optional
            if True, create new sub-thread for function, by default False

        Raises
        ------
        TypeError
            raised if function is not callable
        """

        if not callable(function):
            raise TypeError("function argument cannot be called")
        if not isinstance(args, tuple):
            args = (args,)
        new_event = Event(timestamp, function, args, spawn)
        with self.lock:
            self.event_list.append(new_event)
            evlen = len(self.event_list)
            if not self.onset_idx.any():
                idx = 0
            else:
                idx = np.searchsorted(self.onset_idx[:, 0], timestamp)
            self.onset_idx = np.insert(
                self.onset_idx, idx, [timestamp, evlen - 1], axis=0
            )

    def get(self) -> Event:
        """Get latest event from queue and remove event

        Returns
        -------
        Event
            Latest event
        """
        event = self.peek()
        self.pop()
        return event

    def peek(self) -> Event:
        """Look up latest event from queue

        Returns
        -------
        Event
            Latest event
        """
        with self.lock:
            return self.event_list[int(self.onset_idx[0][1])]

    def empty(self) -> bool:
        """Checks if queue is empty

        Returns
        -------
        bool
            True if queue if empty
        """
        with self.lock:
            return bool(self.event_list)

    def pop(self) -> None:
        """Removes latest event from queue"""
        with self.lock:
            event_idx = int(self.onset_idx[0][1])
            self.onset_idx = self.onset_idx[1:]
            # remove 1 from all idcs after popped event
            self.onset_idx[:, 1][self.onset_idx[:, 1] > event_idx] -= 1
            del self.event_list[event_idx]

    def __worker(self, sleep_time: float, close_event: threading.Event) -> NoReturn:
        """Worker function to process events"""
        while True:
            if close_event.is_set():
                break
            if self.event_list:
                event = self.peek()
                if event.timestamp <= time.time() - self.start:
                    # execute only if not too old
                    if event.timestamp > time.time() - self.start - self.drop_time_thr:
                        event.execute()
                    self.pop()
                # sleep_time = event_list[0].timestamp - (time.time() - self.start) - 0.001
            time.sleep(sleep_time)

    def __repr__(self):
        return f"<TimedQueue {self.event_list.__repr__()}>"

    def elapse(self, time_delta: float) -> None:
        """Add time delta to the current queue time.

        Parameters
        ----------
        time_delta : float
            Additional time
        """
        self.start += time_delta


class TimedQueueSC(TimedQueue):
    """Timed queue with OSC communication.

    Parameters
    ----------
    server : OSCCommunication, optional
        OSC server to handle the bundlers and messsages, by default None
    relative_time : bool, optional
        If True, use relative time, by default False
    thread_sleep_time : float, optional
        Sleep time in seconds for worker thread, by default 0.001
    """

    def __init__(
        self,
        server: OSCCommunication = None,
        relative_time: bool = False,
        thread_sleep_time: float = 0.001,
    ):
        super().__init__(relative_time, thread_sleep_time)
        self.server = server or sc3nb.SC.get_default().server

    def put_bundler(self, onset: float, bundler: Bundler) -> None:
        """Add a Bundler to queue

        Parameters
        ----------
        onset : float
            Sending timetag of the Bundler
        bundler : Bundler
            Bundler that will be sent
        """
        self.put(onset, bundler.send)

    def put_msg(
        self, onset: float, msg: Union[OSCMessage, str], msg_params: Iterable[Any]
    ) -> None:
        """Add a message to queue

        Parameters
        ----------
        onset : float
            Sending timetag of the message
        msg : Union[OSCMessage, str]
            OSCMessage or OSC address
        msg_params : Iterable[Any]
            If msg is str, this will be the parameters of the created OSCMessage
        """
        if isinstance(msg, str):
            self.put(onset, self.server.msg, args=(msg, msg_params))
        else:
            self.put(onset, self.server.send, args=(msg,))



import time
import warnings

from sc3nb.sc_objects.node import Synth, SynthInfo
from tests.conftest import SCBaseTest


class SynthTest(SCBaseTest):
    __test__ = True

    def setUp(self) -> None:
        with self.assertRaises(RuntimeError):
            SynthTest.sc.lang
        self.custom_nodeid = 42
        self.synth_args = {"amp": 0.0, "num": 3}
        warnings.simplefilter("always", UserWarning)
        with self.assertWarnsRegex(
            UserWarning, "SynthDesc 's2' is unknown", msg="SynthDesc seems to be known"
        ):
            self.synth = Synth(
                "s2", controls=self.synth_args, nodeid=self.custom_nodeid
            )
        self.assertIsNone(self.synth._synth_desc)
        self.sc.server.sync()

    def tearDown(self) -> None:
        nodeid = self.synth.nodeid
        self.assertIn(nodeid, self.sc.server.nodes)
        self.synth.free()
        self.synth.wait()
        del self.synth  # make sure that synth is deleted from registry
        t0 = time.time()
        while nodeid in self.sc.server.nodes:
            time.sleep(0.005)
            if time.time() - t0 > 0.2:
                self.fail("NodeID is still in server.nodes")
        self.assertNotIn(nodeid, self.sc.server.nodes)
        with self.assertRaises(KeyError):
            del self.sc.server.nodes[nodeid]

    def test_node_registry(self):
        copy1 = Synth(nodeid=self.synth.nodeid, new=False)
        copy2 = Synth(nodeid=self.custom_nodeid, new=False)
        self.assertIs(self.synth, copy1)
        self.assertIs(self.synth, copy2)
        self.assertIs(copy1, copy2)
        del copy1, copy2

    def test_set_get(self):
        for name, value in {"amp": 0.0, "num": 1}.items():
            self.synth.__setattr__(name, value)
            self.assertAlmostEqual(self.synth.__getattr__(name), value)

        with self.assertWarnsRegex(UserWarning, "Setting 'freq' as python attribute"):
            with self.assertRaisesRegex(AttributeError, "no attribute 'freq'"):
                self.synth.__getattribute__(
                    "freq"
                )  # should not have a python attribute named freq
            self.synth.freq = 420  # should warn if setting attribute

        self.assertAlmostEqual(
            self.synth.get("freq"), 400
        )  # default freq of s2 SynthDef

        with self.assertWarnsRegex(UserWarning, "recognized as Node Parameter now"):
            self.synth.set("freq", 100)
        self.assertAlmostEqual(self.synth.get("freq"), 100)

        self.synth.freq = 300
        self.assertAlmostEqual(self.synth.get("freq"), 300)

        for name, value in self.synth_args.items():
            self.assertAlmostEqual(self.synth.__getattr__(name), value)

    def test_new_warning(self):
        with self.assertLogs(level="WARNING") as log:
            self.synth.new()
            time.sleep(0.1)
        self.assertTrue("duplicate node ID" in log.output[-1])

    def test_query(self):
        query_result = self.synth.query()
        self.assertIsInstance(query_result, SynthInfo)
        self.assertEqual(query_result.nodeid, self.custom_nodeid)
        self.assertEqual(query_result.group, SynthTest.sc.server.default_group.nodeid)
        self.assertEqual(query_result.prev_nodeid, -1)
        self.assertEqual(query_result.next_nodeid, -1)



"""OSC communication

Classes and functions to communicate with SuperCollider
using the Open Sound Control (OSC) protocol over UDP
"""

import atexit
import contextlib
import copy
import errno
import logging
import socket
import threading
import time
import traceback
import warnings
from abc import ABC, abstractmethod
from queue import Empty, Queue
from threading import RLock
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from pythonosc.dispatcher import Dispatcher
from pythonosc.osc_bundle import OscBundle
from pythonosc.osc_bundle_builder import BuildError, OscBundleBuilder
from pythonosc.osc_message import OscMessage
from pythonosc.osc_message_builder import OscMessageBuilder
from pythonosc.osc_packet import OscPacket, ParseError
from pythonosc.osc_server import OSCUDPServer, ThreadingOSCUDPServer
from pythonosc.parsing import osc_types

import sc3nb

_LOGGER = logging.getLogger(__name__)


def get_max_udp_packet_size():
    """Get the max UDP packet size by trial and error"""

    def try_packet_of_size(sock, packet_size):
        address = "127.0.0.1", 50005
        with contextlib.suppress(OSError):
            msg = bytes(1) * packet_size
            if sock.sendto(msg, address) == len(msg):
                return True
        return False

    def recurse_get_max_udp_packet_size(sock, largest_good_size, smallest_bad_size):
        if (largest_good_size + 1) == smallest_bad_size:
            return largest_good_size
        else:
            new_mid_size = int((largest_good_size + smallest_bad_size) / 2)
            if try_packet_of_size(sock, new_mid_size):
                return recurse_get_max_udp_packet_size(
                    sock, new_mid_size, smallest_bad_size
                )
            else:
                return recurse_get_max_udp_packet_size(
                    sock, largest_good_size, new_mid_size
                )

    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        max_udp_packet_size = recurse_get_max_udp_packet_size(s, 0, 65536)
    return max_udp_packet_size


def split_into_max_size(bundler: "Bundler", max_dgram_size) -> List[OscBundle]:
    splits = []
    for timetag, msgs in bundler.messages().items():
        try:
            header = b"#bundle\x00" + osc_types.write_date(timetag)
            bundle_dgram = copy.deepcopy(header)
            for msg in msgs:
                pythonosc_msg = msg.to_pythonosc()
                size = pythonosc_msg.size
                msg_dgram = osc_types.write_int(size)
                msg_dgram += pythonosc_msg.dgram
                if len(bundle_dgram) + len(msg_dgram) > max_dgram_size:
                    splits.append(OscBundle(bundle_dgram))
                    bundle_dgram = copy.deepcopy(header)
                bundle_dgram += msg_dgram
            splits.append(OscBundle(bundle_dgram))
        except osc_types.BuildError as build_error:
            raise BuildError("Could not build the bundle {}".format(build_error))
    return splits


class OSCMessage:
    """Class for creating messages to send over OSC

    Parameters
    ----------
    msg_address : str
        OSC message address
    msg_parameters : Optional[Union[Sequence]], optional
        OSC message parameters, by default None
    """

    # TODO add reply_address and callback functionality to osc_com
    # reply_address : Optional[str], optional
    #     OSC address of the reply to this message, by default None
    # callback : Callable[..., None], optional
    #     Callback for reply handling, by default None

    #     reply_address: Optional[str] = None,
    #     callback: Callable[..., None] = None,

    # self._reply_addr = reply_address
    #     self._callback = callback

    # @property
    # def reply_address(self) -> Optional[str]:
    #     """OSC message reply address"""
    #     return self._content.address

    # @property
    # def callback(self) -> Optional[Callable[..., None]]:
    #     """OSC message reply handler"""
    #     return self._callback

    def __init__(
        self,
        msg_address: str,
        msg_parameters: Optional[Union[Sequence, Any]] = None,
    ) -> None:
        self._content: OscMessage = OSCMessage._build_message(
            msg_address, msg_parameters
        )

    @property
    def dgram(self) -> bytes:
        """datagram of OSC message"""
        return self._content.dgram

    @property
    def raw_osc(self) -> bytes:
        """raw OSC representation - same as :py:attr:`~dgram`"""
        return self.dgram

    @property
    def parameters(self) -> List[Any]:
        """OSC message parameters"""
        return self._content.params

    @property
    def address(self) -> str:
        """OSC message address"""
        return self._content.address

    def to_pythonosc(self) -> OscMessage:
        """Return python-osc OscMessage"""
        return self._content

    @staticmethod
    def _build_message(
        msg_address: str, msg_parameters: Optional[Union[Sequence, Any]] = None
    ) -> OscMessage:
        """Builds pythonsosc OSC message.

        Parameters
        ----------
        msg_address : str
            SuperCollider address.
        msg_parameters : list, optional
            List of parameters to add to message.

        Returns
        -------
        OscMessage
            Message ready to be sent.

        """
        if msg_parameters is None:
            msg_parameters = []
        elif not isinstance(msg_parameters, Sequence) or isinstance(
            msg_parameters, (str, bytes)
        ):
            msg_parameters = [msg_parameters]

        if not msg_address.startswith("/"):
            msg_address = "/" + msg_address

        builder = OscMessageBuilder(address=msg_address)
        for msg_arg in msg_parameters:
            if isinstance(msg_arg, np.number):
                msg_arg = msg_arg.item()
            builder.add_arg(msg_arg)
        return builder.build()

    def __repr__(self) -> str:
        return f'<OSCMessage("{self.address}", {self.parameters})>'


class Bundler:
    """Class for creating OSCBundles and bundling of messages"""

    def __init__(
        self,
        timetag: float = 0,
        msg: Optional[Union[OSCMessage, str]] = None,
        msg_params: Optional[Sequence[Any]] = None,
        *,
        server: Optional["OSCCommunication"] = None,
        receiver: Optional[Union[str, Tuple[str, int]]] = None,
        send_on_exit: bool = True,
    ) -> None:
        """Create a Bundler

        Parameters
        ----------
        timetag : float, optional
            Starting time at which bundle content should be executed.
            If timetag > 1e6 it is interpreted as POSIX time.
            If timetag <= 1e6 it is assumed to be relative value in seconds
            and is added to time.time(), by default 0, i.e. 'now'.
        msg : OSCMessage or str, optional
            OSCMessage or message address, by default None
        msg_params : sequence of any type, optional
            Parameters for the message, by default None
        server : OSCCommunication, optional
            OSC server, by default None
        receiver : Union[str, Tuple[str, int]], optional
            Where to send the bundle, by default send to default receiver of server
        send_on_exit : bool, optional
            Whether the bundle is sent when using as context manager, by default True
        """
        self.timetag = timetag
        self.default_receiver = receiver
        if server is not None:
            self.server = server
        else:
            try:
                self.server = sc3nb.SC.get_default().server
            except RuntimeError:
                self.server = None
        self.contents: List[Union["Bundler", OSCMessage]] = []
        self.passed_time = 0.0
        if msg:
            if not isinstance(msg, OSCMessage):
                msg = OSCMessage(msg, msg_params)
            self.contents.append(msg)
        self.send_on_exit = send_on_exit

    @property
    def dgram(self) -> bytes:
        # Bundler needs to be build, this ensures that
        # relative timings are calculated just now
        return self.to_pythonosc().dgram

    def wait(self, time_passed: float) -> None:
        """Add time to internal time

        Parameters
        ----------
        time_passed : float
            How much seconds should be passed.
        """
        self.passed_time += time_passed

    def add(self, *args) -> "Bundler":
        """Add content to this Bundler.

        Parameters
        ----------
        args : accepts an OSCMessage or Bundler
               or a timetag with an OSCMessage or Bundler
               or Bundler arguments like
                (timetag, msg_addr, msg_params)
                (timetag, msg_addr)
                (timetag, msg)

        Returns
        -------
        Bundler
            self for chaining
        """
        # base case with one argument
        if len(args) == 1:  # args = Message / bundler
            content = args[0]
            if isinstance(content, OSCMessage):
                bundler = Bundler(self.passed_time, content)
            elif isinstance(content, Bundler):
                bundler = copy.deepcopy(content)
                if bundler.timetag < 1e6:
                    bundler.timetag += self.passed_time
            else:
                raise ValueError(
                    f"Cannot add {content} of type {type(content)}. "
                    f"Needing {OSCMessage} or {Bundler} if len(args)==1"
                )
            _LOGGER.debug(
                "%s Appending %s to Bundler with time %s",
                self.passed_time,
                bundler,
                bundler.timetag,
            )
            self.contents.append(bundler)
            return self
        # transform args into Bundler and recurse (return self.add(...))
        elif len(args) == 3:  # args = Bundler args
            timetag, msg_addr, msg_params = args
            return self.add(Bundler(timetag, msg_addr, msg_params))
        elif len(args) == 2:  # args = timetag + Message / Bundler
            timetag, content = args
            if isinstance(content, (OSCMessage, str)):
                return self.add(Bundler(timetag, content))
            if isinstance(content, Bundler):
                bundler = copy.deepcopy(content)
                if bundler.timetag < 1e6:
                    bundler.timetag += timetag
                else:
                    bundler.timetag = timetag
                    warnings.warn(
                        f"Bundler absolute timetag has been overwritten with {timetag}"
                    )
                return self.add(bundler)
        raise ValueError(f"Invalid parameters {args}")

    def messages(
        self, start_time: Optional[float] = 0.0, delay: Optional[float] = None
    ) -> Dict[float, List[OSCMessage]]:
        """Generate a dict with all messages in this Bundler.

        They dict key is the time tag of the messages.

        Parameters
        ----------
        start_time : Optional[float], optional
            start time when using relative timing, by default 0.0

        Returns
        -------
        Dict[float, List[OSCMessage]]
            dict containg all OSCMessages
        """
        start_time = self._calc_timetag(start_time)
        messages = {}
        for content in self.contents:
            if isinstance(content, Bundler):
                for timetag, cont in content.messages(start_time, delay).items():
                    messages.setdefault(timetag, [])
                    messages[timetag].extend(cont)
            elif isinstance(content, OSCMessage):
                timetag = start_time + (delay if delay is not None else 0)
                messages.setdefault(timetag, [])
                messages[timetag].append(content)
        return messages

    def send(
        self,
        server: Optional["OSCCommunication"] = None,
        receiver: Tuple[str, int] = None,
        bundle: bool = True,
    ):
        """Send this Bundler.

        Parameters
        ----------
        server : OSCCommunication, optional
            Server instance for sending the bundle.
            If None it will use the server from init
            or try to use sc3nb.SC.get_default().server, by default None
        receiver : Tuple[str, int], optional
            Address (ip, port) to send to, if None it will send the bundle to
            the default receiver of the Bundler
        bundle : bool, optional
            If True this is allowed to be bundled, by default True

        Raises
        ------
        RuntimeError
            When no server could be found.
        """
        if not server:
            server = self.server or sc3nb.SC.get_default().server
        else:
            raise RuntimeError("No server for sending provided.")
        if receiver is None and self.default_receiver is not None:
            receiver = server.lookup_receiver(self.default_receiver)
        server.send(self, receiver=receiver, bundle=bundle)

    def to_raw_osc(
        self, start_time: Optional[float] = None, delay: Optional[float] = None
    ) -> bytes:
        """Create a raw OSC Bundle from this bundler.

        Parameters
        ----------
        start_time : Optional[float], optional
            used as start time when using relative timing, by default time.time()
        delay: float, optinal
            used to delay the timing.

        Returns
        -------
        OscBundle
            bundle instance for sending
        """
        return self.to_pythonosc(start_time, delay).dgram

    def to_pythonosc(
        self, start_time: Optional[float] = None, delay: Optional[float] = None
    ) -> OscBundle:
        """Build this bundle.

        Parameters
        ----------
        start_time : Optional[float], optional
            used as start time when using relative timing, by default time.time()
        delay: float, optinal
            used to delay the timing.

        Returns
        -------
        OscBundle
            bundle instance for sending
        """
        start_time = self._calc_timetag(start_time)
        # build bundle
        builder = OscBundleBuilder(start_time + (delay if delay is not None else 0))
        # add contents
        for content in self.contents:
            if isinstance(content, Bundler):
                builder.add_content(
                    content.to_pythonosc(start_time=start_time, delay=delay)
                )
            elif isinstance(content, OSCMessage):
                builder.add_content(content.to_pythonosc())
            else:
                ValueError("Couldn't build with unsupported content: {content}")
        return builder.build()

    def _calc_timetag(self, start_time: Optional[float]):
        if self.timetag > 1e6:
            # absolute time
            return self.timetag
        # use relativ timing
        # if start_time is None use just now
        if start_time is None:
            start_time = time.time()
        # time tag = relative offset (start time) + relative timetag
        return start_time + self.timetag

    def __deepcopy__(self, memo) -> "Bundler":
        new_bundler = Bundler(self.timetag, server=self.server)
        new_bundler.contents = copy.deepcopy(self.contents)
        return new_bundler

    def __enter__(self):
        self.server._bundling_lock.acquire(timeout=1)
        # if there is already a bundling bundle set when we have
        # the lock it is from the same thread.
        self.server._bundling_bundles.append(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.server._bundling_bundles.pop() is not self:
            raise RuntimeError("Bundler nesting failed.")
        self.server._bundling_lock.release()
        if exc_value is not None:
            raise RuntimeError(
                "Aborting. Exception raised in bundler:\n"
                f"{''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))}"
            )
        elif self.send_on_exit:
            self.send(bundle=True)

    def __repr__(self) -> str:
        messages_str = repr(self.messages())
        if len(messages_str) > 330:
            messages_str = messages_str[:150] + "  ...  " + messages_str[-150:]
        return f"<Bundler {messages_str}>"


def convert_to_sc3nb_osc(
    data: Union[OSCMessage, Bundler, OscMessage, OscBundle, bytes]
) -> Union[OSCMessage, Bundler]:
    """Get binary OSC representation

    Parameters
    ----------
    package : Union[OscMessage, Bundler, OscBundle]
        OSC Package object

    Returns
    -------
    bytes
        raw OSC binary representation of OSC Package

    Raises
    ------
    ValueError
        If package is not supported
    """
    if isinstance(data, (OSCMessage, Bundler)):
        return data

    if isinstance(data, bytes):
        dgram = data
    else:
        try:
            dgram = data.dgram
            packet = OscPacket(dgram)
        except (AttributeError, ParseError) as error:
            raise ValueError(
                f"Unsupported data of type {type(data)} : {data}"
            ) from error
        if OscMessage.dgram_is_message(dgram) and len(packet.messages) == 1:
            message = packet.messages[0]
            return OSCMessage(message.address, message.params)

        bundler = Bundler()
        for timed_msg in packet.messages:
            bundler.add(
                timed_msg.time,
                OSCMessage(timed_msg.message.address, timed_msg.message.params),
            )
        return bundler


class MessageHandler(ABC):
    """Base class for Message Handling"""

    @property
    @abstractmethod
    def map_values(self) -> Tuple[str, Callable]:
        """Values used to setup mapping

        Returns
        -------
        Tuple[str, Callable]
            OSC address, corresponding callback
        """

    @abstractmethod
    def put(self, address: str, *args) -> None:
        """Add message to MessageHandler

        Parameters
        ----------
        address : str
            Message address
        """


class MessageQueue(MessageHandler):
    """Queue to retrieve OSC messages send to the corresponding OSC address"""

    def __init__(self, address: str, preprocess: Optional[Callable] = None):
        """Create a new AddressQueue

        Parameters
        ----------
        address : str
            OSC address for this queue
        preprocess : function, optional
            function that will be applied to the value before they are enqueued
             (Default value = None)
        """
        self._address = address
        self.process = preprocess
        self._queue = Queue()
        self._skips = 0

    def put(self, address: str, *args) -> None:
        """Add a message to MessageQueue

        Parameters
        ----------
        address : str
            message address
        """
        if self._address != address:
            _LOGGER.warning(
                "AddressQueue %s: alternative address %s", self._address, address
            )
        if self.process:
            args = self.process(args)
        else:
            if len(args) == 1:
                args = args[0]
        self._queue.put(args)

    @property
    def skips(self) -> int:
        """Counts how many times this queue was not synced"""
        return self._skips

    @property
    def size(self) -> int:
        """How many items are in this queue"""
        return self._queue.qsize()

    def skipped(self):
        """Skipp one queue value"""
        self._skips += 1

    @property
    def map_values(self) -> Tuple[str, Callable]:
        """Values needed for dispatcher map call

        Returns
        -------
        tuple
            (OSC address pattern, callback function)
        """
        return self._address, self.put

    def get(self, timeout: float = 5, skip: bool = True) -> Any:
        """Returns a value from the queue

        Parameters
        ----------
        timeout : int, optional
            Time in seconds that will be waited on the queue, by default 5
        skip : bool, optional
            If True the queue will skip as many values as `skips`, by default True

        Returns
        -------
        obj
            value from queue

        Raises
        ------
        Empty
            If the queue has no value

        """
        if skip:
            while self._skips > 0:
                skipped_value = self._queue.get(block=True, timeout=timeout)
                _LOGGER.warning(
                    "AddressQueue %s: skipped value %s", self._address, skipped_value
                )
                self._skips -= 1
        if self._skips > 0:
            self._skips -= 1
        val = self._queue.get(block=True, timeout=timeout)
        self._queue.task_done()
        return val

    def show(self) -> None:
        """Print the content of the queue."""
        print(list(self._queue.queue))

    def _repr_pretty_(self, printer, cycle) -> None:
        if cycle:
            printer.text("AddressQueue")
        else:
            printer.text(f"AddressQueue {self._address} : {list(self._queue.queue)}")


class MessageQueueCollection(MessageHandler):
    """A collection of MessageQueues that are all sent to one and the same first address."""

    def __init__(self, address: str, sub_addrs: Optional[Sequence[str]] = None):
        """Create a collection of MessageQueues under the same first address

        Parameters
        ----------
        address : str
            first message address that is the same for all MessageQueues
        sub_addrs : Optional[Sequence[str]], optional
            secound message addresses with seperate queues, by default None
            Additional MessageQueues will be created on demand.
        """
        self._address = address
        if sub_addrs is not None:
            self.msg_queues = {
                msg_addr: MessageQueue(msg_addr) for msg_addr in sub_addrs
            }
        else:
            self.msg_queues = {}

    def put(self, address: str, *args) -> None:
        """Add a message to the corresponding MessageQueue

        Parameters
        ----------
        address : str
            first message address
        """
        subaddress, *args = args
        if subaddress not in self.msg_queues:
            self.msg_queues[subaddress] = MessageQueue(subaddress)
            _LOGGER.debug(
                "MessageQueue for %s was created under MessageQueueCollection %s.",
                subaddress,
                self._address,
            )
        self.msg_queues[subaddress].put(subaddress, *args)

    @property
    def map_values(self) -> Tuple[str, Callable]:
        """Values needed for dispatcher map call

        Returns
        -------
        tuple
            (OSC address pattern, callback function)
        """
        return self._address, self.put

    def __contains__(self, item) -> bool:
        return item in self.msg_queues

    def __getitem__(self, key):
        return self.msg_queues[key]


class OSCCommunicationError(Exception):
    """Exception for OSCCommunication errors."""

    def __init__(self, message, send_message):
        self.message = message
        self.send_message = send_message
        super().__init__(self.message)


class OSCCommunication:
    """Class to send and receive OSC messages and bundles."""

    def __init__(
        self,
        server_ip: str,
        server_port: int,
        default_receiver_ip: str,
        default_receiver_port: int,
    ) -> None:
        """Create an OSC communication server

        Parameters
        ----------
        server_ip : str
            IP address to use for this server
        server_port : int
            port to use for this server
        default_receiver_ip : str
            IP address used for sending by default
        default_receiver_port : int
            port used for sending by default
        """
        self._receivers: Dict[Tuple[str, int], str] = {}
        self._default_receiver: Tuple[str, int] = (
            default_receiver_ip,
            default_receiver_port,
        )

        # bundling messages support
        self._bundling_lock = RLock()
        self._bundling_bundles = []

        # create server
        while True:
            try:
                self._osc_server = ThreadingOSCUDPServer(
                    (server_ip, server_port), Dispatcher()
                )
                self._osc_server_running = True
                _LOGGER.debug(
                    "This OSCCommunication instance is at port: %s", server_port
                )
                break
            except OSError as error:
                if error.errno == errno.EADDRINUSE:
                    server_port += 1

        # start server thread
        self._osc_server_thread = threading.Thread(
            target=self._osc_server.serve_forever
        )
        self._osc_server_thread.daemon = True
        self._osc_server_thread.start()
        self._max_udp_packet_size = None

        # init queues for msg pairs, must be after self._osc_server
        self._msg_queues: Dict[str, MessageQueue] = {}
        self._reply_addresses: Dict[str, str] = {}

        atexit.register(self.quit)

    @property
    def osc_server(self) -> OSCUDPServer:
        """Underlying OSC server"""
        return self._osc_server

    def add_msg_pairs(self, msg_pairs: Dict[str, str]) -> None:
        """Add the provided pairs for message receiving.

        Parameters
        ----------
        msg_pairs : dict[str, str], optional
            dict containing user specified message pairs.
            {msg_addr: reply_addr}
        """
        for msg_addr, reply_addr in msg_pairs.items():
            self.add_msg_queue(MessageQueue(reply_addr), msg_addr)

    def add_msg_queue(
        self, msg_queue: MessageQueue, out_addr: Optional[str] = None
    ) -> None:
        """Add a MessageQueue to this servers dispatcher

        Parameters
        ----------
        msg_queue : MessageQueue
            new MessageQueue
        out_addr : Optional[str], optional
            The outgoing message address that belongs to this MessageQeue, by default None
        """
        reply_addr, handler = msg_queue.map_values
        if reply_addr in self._msg_queues or out_addr in self._reply_addresses:
            warnings.warn(f"Overwriting handler for ({out_addr} -> {reply_addr})")
        self._osc_server.dispatcher.map(reply_addr, handler)
        self._msg_queues[reply_addr] = msg_queue
        if out_addr:
            self._reply_addresses[out_addr] = reply_addr

    def add_msg_queue_collection(
        self, msg_queue_collection: MessageQueueCollection
    ) -> None:
        """Add a MessageQueueCollection

        Parameters
        ----------
        msg_queue_collection : MessageQueueCollection
            MessageQueueCollection to be added
        """
        collection_addr, handler = msg_queue_collection.map_values
        self._osc_server.dispatcher.map(collection_addr, handler)
        for msg_addr, msg_queue in msg_queue_collection.msg_queues.items():
            self._msg_queues[collection_addr + msg_addr] = msg_queue
            self._reply_addresses[msg_addr] = collection_addr + msg_addr

    @property
    def msg_queues(self) -> Dict[str, MessageQueue]:
        """Dict with all added MessageQueues

        Returns
        -------
        Dict[str, MessageQueue]
            Queue address, MessageQueue pairs
        """
        return self._msg_queues

    @property
    def reply_addresses(self) -> Dict[str, str]:
        """Dict with all addresses and the replies

        Returns
        -------
        Dict[str, str]
            Outgoing address, incoming address
        """
        return self._reply_addresses

    def _check_sender(self, sender: Tuple[str, int]) -> Union[str, Tuple[str, int]]:
        return self._receivers.get(sender, sender)

    def lookup_receiver(self, receiver: Union[str, Tuple[str, int]]) -> Tuple[str, int]:
        """Reverse lookup the address of a specific receiver

        Parameters
        ----------
        receiver : str
            Receiver name.

        Returns
        -------
        Tuple[str, int]
            Receiver address (ip, port)

        Raises
        ------
        KeyError
            If receiver is unknown.
        ValueError
            If the type of the receiver argument is wrong.
        """
        if isinstance(receiver, str):
            try:
                return next(
                    addr for addr, name in self._receivers.items() if name == receiver
                )
            except StopIteration as error:
                raise KeyError from error
        elif isinstance(receiver, tuple):
            return receiver
        else:
            raise ValueError(f"Incorrect type for receiver ({receiver}).")

    def connection_info(
        self, print_info: bool = True
    ) -> Tuple[Tuple[str, int], Dict[Tuple[str, int], str]]:
        """Get information about the known addresses

        Parameters
        ----------
        print_info : bool, optional
            If True print connection information
             (Default value = True)

        Returns
        -------
        tuple
            containing the address of this sc3nb OSC Server
            and known receivers addresses in a dict with their names as values

        """
        if print_info:
            receivers_str = "".join(
                f'"{name}" at {addr}\n                 '
                for addr, name in self._receivers.items()
            )
            print(
                f"This instance is at {self._osc_server.server_address},\n"
                f"Known receivers: {receivers_str}"
            )
        return (self._osc_server.server_address, self._receivers)

    def add_receiver(self, name: str, ip_address: str, port: int):
        """Adds a receiver with the specified address.

        Parameters
        ----------
        name : str
            Name of receiver.
        ip_address : str
            IP address of receiver (e.g. "127.0.0.1")
        port : int
            Port of the receiver
        """
        self._receivers[(ip_address, port)] = name

    def send(
        self,
        package: Union[OSCMessage, Bundler],
        *,
        receiver: Optional[Union[str, Tuple[str, int]]] = None,
        bundle: bool = False,
        await_reply: bool = True,
        timeout: float = 5,
    ) -> Any:
        """Sends OSC packet

        Parameters
        ----------
        package : OSCMessage or Bundler
            Object with `dgram` attribute.
        receiver : str or Tuple[str, int], optional
            Where to send the packet, by default send to default receiver
        bundle : bool, optional
            If True it is allowed to bundle the package with bundling, by default False.
        await_reply : bool, optional
            If True ask for reply from the server and return it,
            otherwise send the message and return None directly, by default True.
            If the package is bundled None will be returned.
        timeout : int, optional
            timeout in seconds for reply, by default 5

        Returns
        -------
        None or reply
            None if no reply was received or awaited else reply.

        Raises
        ------
        ValueError
            When the provided package is not supported.
        OSCCommunicationError
            When the handling of a package fails.
        """
        # TODO we could use a typing.Protocol for sendableOSC (.dgram), ..
        # bundling
        if bundle:
            with self._bundling_lock:
                if self._bundling_bundles:
                    self._bundling_bundles[-1].add(package)
                    return

        if receiver is not None:
            receiver_address = self.lookup_receiver(receiver)
        else:
            receiver_address = self._default_receiver

        package = convert_to_sc3nb_osc(package)
        try:
            sent_bytes = self._osc_server.socket.sendto(package.dgram, receiver_address)
        except OSError as error:
            if self._max_udp_packet_size is None:
                self._max_udp_packet_size = get_max_udp_packet_size()
            if (
                isinstance(package, Bundler)
                and len(package.dgram) > self._max_udp_packet_size
            ):
                _LOGGER.warning(
                    f"OSC Bundle is too large ({len(package.dgram)}/{self._max_udp_packet_size}) and will be splitted."
                )
                osc_bundles = split_into_max_size(package, self._max_udp_packet_size)
                for osc_bundle in osc_bundles:
                    if (
                        self._osc_server.socket.sendto(
                            osc_bundle.dgram, receiver_address
                        )
                        == 0
                    ):
                        raise RuntimeError(
                            "Could not send data. Socket connection broken."
                        )
            else:
                raise OSCCommunicationError(
                    f"Sending OSC package failed - {error}", package
                ) from error
        else:
            if sent_bytes == 0:
                raise RuntimeError("Could not send data. Socket connection broken.")

        if isinstance(package, OSCMessage):
            return self._handle_outgoing_message(
                package, receiver_address, await_reply, timeout
            )
        elif isinstance(package, Bundler):
            # logging
            if _LOGGER.isEnabledFor(logging.INFO):
                _LOGGER.info(
                    "send to %s : %s contents size %s ",
                    self._check_sender(receiver_address),
                    package,
                    len(package.contents),
                )
            # handling
            # for each message we should skip queues here
            for _, msgs in package.messages().items():
                for msg in msgs:
                    self._handle_outgoing_message(msg, receiver_address, False, timeout)
        else:
            _LOGGER.info("send to %s : %s", receiver, package)

    def _handle_outgoing_message(
        self,
        message: OSCMessage,
        receiver_address: Tuple[str, int],
        await_reply: bool,
        timeout: float,
    ) -> Any:
        # logging
        if _LOGGER.isEnabledFor(logging.INFO):
            msg_params_str = str(message.parameters)
            if not _LOGGER.isEnabledFor(logging.DEBUG) and len(msg_params_str) > 55:
                msg_params_str = msg_params_str[:55] + ".."
            _LOGGER.debug(
                "send to %s : %s %s",
                self._check_sender(receiver_address),
                message.address,
                msg_params_str,
            )
        # handling
        reply_addr = self.get_reply_address(message.address)
        try:
            if reply_addr is not None and reply_addr in self._msg_queues:
                if await_reply:
                    return self._msg_queues[reply_addr].get(timeout, skip=True)
                else:
                    self._msg_queues[reply_addr].skipped()
                    return
        except (Empty, TimeoutError) as error:
            if isinstance(error, Empty):
                error_msg = (
                    f"Failed to get reply at '{reply_addr}' "
                    f"after '{message.address}' message to "
                )
            elif isinstance(error, TimeoutError):
                error_msg = f"Timed out after '{message.address}' message to "
            else:
                error_msg = f"Error when sending '{message.address}' message to "
            error_msg += f"{self._check_sender(receiver_address)}"
            raise OSCCommunicationError(error_msg, message) from error

    def get_reply_address(self, msg_address: str) -> Optional[str]:
        """Get the corresponding reply address for the given address

        Parameters
        ----------
        msg_address : str
            outgoing message address

        Returns
        -------
        str or None
            Corresponding reply address if available
        """
        return self._reply_addresses.get(msg_address, None)

    def msg(
        self,
        msg_addr: str,
        msg_params: Optional[Sequence] = None,
        *,
        bundle: bool = False,
        receiver: Optional[Tuple[str, int]] = None,
        await_reply: bool = True,
        timeout: float = 5,
    ) -> Optional[Any]:
        """Creates and sends OSC message over UDP.

        Parameters
        ----------
        msg_addr : str
            SuperCollider address of the OSC message
        msg_params : Optional[Sequence], optional
            List of paramters of the OSC message, by default None
        bundle : bool, optional
            If True it is allowed to bundle the content with bundling, by default False
        receiver : tuple[str, int], optional
            (IP address, port) to send the message, by default send to default receiver
        await_reply : bool, optional
            If True send message and wait for reply
            otherwise send the message and return directly, by default True
        timeout : float, optional
            timeout in seconds for reply, by default 5

        Returns
        -------
        obj
            reply if await_reply and there is a reply for this
        """
        return self.send(
            OSCMessage(msg_addr, msg_params),
            bundle=bundle,
            receiver=receiver,
            await_reply=await_reply,
            timeout=timeout,
        )

    def bundler(
        self,
        timetag: float = 0,
        msg: Optional[Union[OSCMessage, str]] = None,
        msg_params: Optional[Sequence[Any]] = None,
        send_on_exit: bool = True,
    ) -> Bundler:
        """Generate a Bundler.

        This allows the user to easly add messages/bundles and send it.

        Parameters
        ----------
        timetag : int
            Time at which bundle content should be executed.
            If timetag <= 1e6 it is added to time.time().
        msg : OSCMessage or str, optional
            OSCMessage or message address, by default None
        msg_params : sequence of any type, optional
            Parameters for the message, by default None
        send_on_exit : bool, optional
            Whether the bundle is sent when using as context manager, by default True

        Returns
        -------
        Bundler
            bundler for OSC bundling.
        """
        return Bundler(
            timetag=timetag,
            msg=msg,
            msg_params=msg_params,
            server=self,
            send_on_exit=send_on_exit,
        )

    def quit(self) -> None:
        """Shuts down the sc3nb OSC server"""
        if self._osc_server_running:
            self._osc_server.shutdown()
            self._osc_server.server_close()
            self._osc_server_thread.join(timeout=5)
            self._osc_server_running = False



"""Module for parsing OSC packets from sclang.

This implements an extension of the OSC protocol.
A bundle is now allowed to consist of other bundles or
lists.

This extension is needed as sclang is sending Arrays
as this list or when nested as bundles with inner list

"""
import logging
import math
from typing import Any, Sequence, Tuple, Union

from pythonosc.osc_bundle import OscBundle
from pythonosc.parsing import osc_types

_LOGGER = logging.getLogger(__name__)


SYNTH_DEF_MARKER = b"SCgf"
TYPE_TAG_MARKER = ord(b",")
TYPE_TAG_INDEX = 4
NUM_SIZE = 4
BYTES_2_TYPE = {
    "i": osc_types.get_int,
    "f": osc_types.get_float,
    "d": osc_types.get_double,
    "s": osc_types.get_string,
    "N": lambda dgram, start_index: (None, start_index),
    "I": lambda dgram, start_index: (float("inf"), start_index),
    "T": lambda dgram, start_index: (True, start_index),
    "F": lambda dgram, start_index: (False, start_index),
}


class ParseError(Exception):
    """Base exception for when a datagram parsing error occurs."""


def _get_aligned_index(index: int) -> int:
    """Get next multiple of NUM_SIZE from index

    Parameters
    ----------
    index : int
        starting index

    Returns
    -------
    int
        next multiple of NUM_SIZE from index
    """
    return NUM_SIZE * math.ceil(index / NUM_SIZE)


def _parse_list(dgram: bytes, start_index: int) -> Tuple[Sequence[Any], int]:
    """Parse a OSC List

    List consists of the following bytes:
    4 bytes (int) : list_size
    n bytes (string) : OSC type tag
    n bytes (x) : content as specified by type tag

    Parameters
    ----------
    dgram : bytes
        datagram with the list
    start_index : int
        parsing starting index

    Returns
    -------
    Tuple[Sequence[Any], int]
        parsed list contents, starting index + number of consumed bytes

    Raises
    ------
    ParseError
        If datagram is invalid.
    """
    # parse list size
    _LOGGER.debug("[ start parsing list: %s", dgram[start_index:])
    list_size, start_index = osc_types.get_int(dgram, start_index)

    # parse type tag
    type_tag, start_index = osc_types.get_string(dgram, start_index)
    if type_tag.startswith(","):
        type_tag = type_tag[1:]

    _LOGGER.debug("list with size %d and content '%s'", list_size, type_tag)

    # parse content
    value_list = []
    for tag in type_tag:
        try:
            value, start_index = BYTES_2_TYPE[tag](dgram, start_index)
        except KeyError:
            raise ParseError('type tag "{}" not understood'.format(chr(tag)))
        _LOGGER.debug("new value %s", value)
        value_list.append(value)

    _LOGGER.debug("resulting list %s", value_list)
    _LOGGER.debug("] end parsing list")
    return value_list, start_index


def _parse_osc_bundle_element(
    dgram: bytes, start_index: int
) -> Tuple[Union[Sequence[Any], bytes], int]:
    """Parse an element from an OSC bundle.

    The element needs to be either an OSC bundle or a list

    Parameters
    ----------
    dgram : bytes
        datagram with the bundle element
    start_index : int
        parsing starting index

    Returns
    -------
    Tuple[Union[Sequence[Any], bytes], int]
        parsed content of the bundle element, starting index + number of consumed bytes

    Raises
    ------
    ParseError
        If the datagram is invalid.
    """
    elem_size, start_index = osc_types.get_int(dgram, start_index)
    _LOGGER.debug(
        ">> parse OSC bundle element (size: %d): %s ",
        elem_size,
        dgram[start_index : start_index + elem_size],
    )

    if OscBundle.dgram_is_bundle(dgram[start_index : start_index + elem_size]):
        _LOGGER.debug("found bundle")
        msgs, start_index = _parse_bundle(dgram, start_index)
        return msgs, start_index

    if dgram[start_index + TYPE_TAG_INDEX] == TYPE_TAG_MARKER:
        _LOGGER.debug("found list")
        value_list, start_index = _parse_list(dgram, start_index)
        return value_list, start_index

    if dgram[start_index : start_index + 4] == SYNTH_DEF_MARKER:
        _LOGGER.debug("found SynthDef blob")
        synth_def = dgram[start_index : start_index + elem_size]
        start_index = start_index + elem_size
        return synth_def, start_index

    raise ParseError("Datagram not recognized")


BYTES_2_TYPE["b"] = _parse_osc_bundle_element


def _parse_bundle(dgram: bytes, start_index: int) -> Tuple[Sequence[Any], int]:
    """Parsing bundle

    Parameters
    ----------
    dgram : bytes
        datagram with the bundle
    start_index : int
        parsing starting index

    Returns
    -------
    tuple[Sequence[Any], int]
        parsed content, starting index + number of consumed bytes

    Raises
    ------
    ParseError
        If the datagram is invalid
    """
    _LOGGER.debug("## start parsing bundle: %s", dgram[start_index:])

    if dgram[start_index : start_index + 8] != b"#bundle\x00":
        raise ParseError("Datagram of bundles should start with b'#bundle\x00'")
    start_index += 8

    timetag, start_index = osc_types.get_timetag(dgram, start_index)
    _LOGGER.debug("bundle timetag: %s", timetag)

    msgs = []
    while start_index < len(dgram):
        sc_msg, start_index = _parse_osc_bundle_element(dgram, start_index)
        msgs.append(sc_msg)

    start_index = _get_aligned_index(start_index)
    _LOGGER.debug("parsed bytes %s", dgram[:start_index])
    _LOGGER.debug("msgs %s", msgs)
    _LOGGER.debug("## end parsing bundle ")
    return msgs, start_index


def parse_sclang_osc_packet(data: bytes) -> Union[bytes, Sequence[Any]]:
    """Parses the OSC packet from sclang.

    Parameters
    ----------
    data : bytes
        bytes sent by sclang

    Returns
    -------
    bytes or Sequence[Any]
        unchanged bytes or content of bundles/messages
    """
    try:
        if len(data) > TYPE_TAG_INDEX + 1:
            if data[TYPE_TAG_INDEX] == TYPE_TAG_MARKER:
                return _parse_list(data, 0)[0]
            elif data[:8] == b"#bundle\x00":
                return _parse_bundle(data, 0)[0]
    except ParseError as error:
        _LOGGER.warning("Ignoring ParseError:\n%s\nreturning blob", error)
    return data


def preprocess_return(value: Sequence[Any]) -> Sequence[Any]:
    """Preprocessing function for /return values

    Parameters
    ----------
    value : tuple
        return data

    Returns
    -------
    obj
        data

    """
    if len(value) == 1:
        value = value[0]
        if isinstance(value, bytes):
            value = parse_sclang_osc_packet(value)
    return value



import logging
from pathlib import Path
from random import randrange
from unittest import TestCase

import pytest

from sc3nb.sc import startup
from sc3nb.sc_objects.server import ServerOptions


@pytest.fixture(autouse=True)
def no_logs_gte_error(caplog, request):
    yield  # Yielding here runs the test itself
    if "allowloggingwarn" in request.keywords:
        return
    errors = [
        record
        for record in caplog.get_records("call")
        if record.levelno >= logging.WARNING
        and "pythonosc/osc_bundle.py" not in Path(record.pathname).as_posix()
    ]
    formatter = logging.Formatter("%(name)s:%(lineno)d  %(levelname)s - %(message)s")
    assert (
        not errors
    ), f"Got Warning logged: {[formatter.format(error) for error in errors]}"


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "allowloggingwarn: mark for allowing logger warnings"
    )


class SCBaseTest(TestCase):
    __test__ = False
    sc = None
    start_sclang = False

    @classmethod
    def setUpClass(cls) -> None:
        cls.sc = startup(
            start_server=True,
            scsynth_options=ServerOptions(
                udp_port=57777,  # randrange(57777, 58888), # can be useful for debugging tests
                max_logins=3,
            ),
            with_blip=False,
            start_sclang=cls.start_sclang,
        )
        cls.sc.server.dump_osc(1)
        # ensure we get all warnings
        cls.sc.server._max_warning_freq = float("-inf")
        assert cls.sc.server.sync(), "Syncing scsynth failed"
        if cls.start_sclang:
            cls.sc.server.mute()
            assert cls.sc.lang.started, "Starting sclang failed"

    @classmethod
    def tearDownClass(cls) -> None:
        cls.sc.exit()



"""Module for using SuperCollider Buffers in Python"""

import os
import warnings
from enum import Enum, unique
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any, List, NamedTuple, Optional, Sequence, Union

import numpy as np
import scipy.io.wavfile as wavfile

import sc3nb
from sc3nb.sc_objects.node import Synth
from sc3nb.sc_objects.synthdef import SynthDef

if TYPE_CHECKING:
    import pya

    from sc3nb.sc_objects.server import SCServer


@unique
class BufferReply(str, Enum):
    """Buffer Command Replies"""

    INFO = "/b_info"


@unique
class BufferCommand(str, Enum):
    """Buffer OSC Commands for Buffers"""

    ALLOC = "/b_alloc"
    ALLOC_READ = "/b_allocRead"
    ALLOC_READ_CHANNEL = "/b_allocReadChannel"
    READ = "/b_read"
    READ_CHANNEL = "/b_readChannel"
    WRITE = "/b_write"
    FREE = "/b_free"
    ZERO = "/b_zero"
    SET = "/b_set"
    SETN = "/b_setn"
    FILL = "/b_fill"
    GEN = "/b_gen"
    CLOSE = "/b_close"
    QUERY = "/b_query"
    GET = "/b_get"
    GETN = "/b_getn"


@unique
class BufferAllocationMode(str, Enum):
    """Buffer Allocation Modes"""

    FILE = "file"
    ALLOC = "alloc"
    DATA = "data"
    EXISTING = "existing"
    COPY = "copy"
    NONE = "none"


class BufferInfo(NamedTuple):
    """Information about the Buffer"""

    bufnum: int
    num_frames: int
    num_channels: int
    sample_rate: float


class Buffer:
    """A Buffer object represents a SuperCollider3 Buffer on scsynth
    and provides access to low-level buffer commands of scsynth via
    methods of the Buffer objects.

    The constructor merely initializes a buffer:

    * it selects a buffer number using the server's buffer allocator
    * it initializes attribute variables

    Parameters
    ----------
    bufnum : int, optional
        buffer number to be used on scsynth. Defaults to None,
        can be set to enforce a given bufnum
    server : SCServer, optional
        The server instance to establish the Buffer,
        by default use the SC default server

    Attributes
    ----------
    server : the SCServer object
        to communicate with scsynth
    _bufnum : int
        buffer number = bufnum id on scsynth
    _sr : int
        the sampling rate of the buffer
    _channels : int
        number of channels of the buffer
    _samples : int
        buffer length = number of sample frames
    _alloc_mode : str
        ['file', 'alloc', 'data', 'existing', 'copy']
        according to previously used generator, defaults to None
    _allocated : boolean
        True if Buffer has been allocated by
        any of the initialization methods
    _path : str
        path to the audio file used in load_file()

    Notes
    -----
    For more information on Buffer commands, refer to the Server Command Reference in SC3.
    https://doc.sccode.org/Reference/Server-Command-Reference.html#Buffer%20Commands

    Examples
    --------
    (see examples/buffer-examples.ipynb)

    >>> b = Buffer().read(...)
    >>> b = Buffer().load_data(...)
    >>> b = Buffer().alloc(...)
    >>> b = Buffer().load_asig(...)
    >>> b = Buffer().use_existing(...)
    >>> b = Buffer().copy(Buffer)

    """

    def __init__(
        self, bufnum: Optional[int] = None, server: Optional["SCServer"] = None
    ) -> None:
        self._server = server or sc3nb.SC.get_default().server
        self._bufnum_set_manually = bufnum is not None
        self._bufnum = bufnum
        self._sr = None
        self._channels = None
        self._samples = None
        self._alloc_mode = BufferAllocationMode.NONE
        self._allocated = False
        self._path = None
        self._synth_def = None
        self._synth = None

    # Section: Buffer initialization methods
    def read(
        self,
        path: str,
        starting_frame: int = 0,
        num_frames: int = -1,
        channels: Optional[Union[int, Sequence[int]]] = None,
    ) -> "Buffer":
        """Allocate buffer memory and read a sound file.

        If the number of frames argument num_frames is negative or zero,
        the entire file is read.

        Parameters
        ----------
        path : string
            path name of a sound file.
        starting_frame : int
            starting frame in file
        num_frames : int
            number of frames to read
        channels : list | int
            channels and order of channels to be read from file.
            if only a int is provided it is loaded as only channel

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is already allocated.
        """
        if self._allocated:
            raise RuntimeError("Buffer object is already initialized!")
        if self._bufnum is None:
            self._bufnum = self._server.buffer_ids.allocate(num=1)[0]
        self._alloc_mode = BufferAllocationMode.FILE
        self._path = Path(path).resolve(strict=True)
        self._sr, data = wavfile.read(
            self._path
        )  # TODO: we only need the metadata here
        server_sr = self._server.nominal_sr
        if self._sr != server_sr:
            warnings.warn(
                f"Sample rate of file ({self._sr}) does not "
                f"match the SC Server sample rate ({server_sr})"
            )
        self._samples = data.shape[0] if num_frames <= 0 else num_frames
        if channels is None:
            channels = [0] if len(data.shape) == 1 else range(data.shape[1])
        elif isinstance(channels, int):
            channels = [channels]
        self._channels = len(channels)
        self._server.msg(
            BufferCommand.ALLOC_READ_CHANNEL,
            [self._bufnum, str(self._path), starting_frame, num_frames, *channels],
            bundle=True,
        )
        self._allocated = True
        return self

    def alloc(self, size: int, sr: int = 44100, channels: int = 1) -> "Buffer":
        """Allocate buffer memory.

        Parameters
        ----------
        size : int
            number of frames
        sr : int
            sampling rate in Hz (optional. default = 44100)
        channels : int
            number of channels (optional. default = 1 channel)

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is already allocated.
        """
        if self._allocated:
            raise RuntimeError("Buffer object is already initialized!")
        if self._bufnum is None:
            self._bufnum = self._server.buffer_ids.allocate(num=1)[0]
        self._sr = sr
        self._alloc_mode = BufferAllocationMode.ALLOC
        self._channels = channels
        self._samples = int(size)
        self._server.msg(
            BufferCommand.ALLOC, [self._bufnum, size, channels], bundle=True
        )
        self._allocated = True
        return self

    def load_data(
        self,
        data: np.ndarray,
        sr: int = 44100,
        mode: str = "file",
        sync: bool = True,
    ) -> "Buffer":
        """Allocate buffer memory and read input data.

        Parameters
        ----------
        data : numpy array
            Data which should inserted
        sr : int, default: 44100
            sample rate
        mode : 'file' or 'osc'
            Insert data via filemode ('file') or n_set OSC commands ('osc')
            Bundling is only supported for 'osc' mode and if sync is False.
        sync: bool, default: True
            Use SCServer.sync after sending messages when mode = 'osc'

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is already allocated.
        """
        if self._allocated:
            raise RuntimeError("Buffer object is already initialized!")
        if self._bufnum is None:
            self._bufnum = self._server.buffer_ids.allocate(num=1)[0]
        self._alloc_mode = BufferAllocationMode.DATA
        self._sr = sr
        self._samples = data.shape[0]
        self._channels = 1 if len(data.shape) == 1 else data.shape[1]
        if mode == "file":
            tempfile = NamedTemporaryFile(delete=False)
            try:
                wavfile.write(tempfile, self._sr, data)
            finally:
                tempfile.close()
            self._server.msg(
                BufferCommand.ALLOC_READ,
                [self._bufnum, tempfile.name],
                await_reply=True,
            )
            if os.path.exists(tempfile.name):
                os.remove(tempfile.name)
        elif mode == "osc":
            self._server.msg(
                BufferCommand.ALLOC, [self._bufnum, data.shape[0]], bundle=True
            )
            blocksize = 1000  # array size compatible with OSC packet size
            # TODO: check how this depends on datagram size
            # TODO: put into Buffer header as const if needed elsewhere...
            if self._channels > 1:
                data = data.reshape(-1, 1)
            if data.shape[0] < blocksize:
                self._server.msg(
                    BufferCommand.SETN,
                    [self._bufnum, [0, data.shape[0], data.tolist()]],
                    bundle=True,
                )
            else:
                # For datasets larger than {blocksize} entries,
                # split data to avoid network problems
                splitdata = np.array_split(data, data.shape[0] / blocksize)
                for i, chunk in enumerate(splitdata):
                    self._server.msg(
                        BufferCommand.SETN,
                        [self._bufnum, i * blocksize, chunk.shape[0], chunk.tolist()],
                        await_reply=False,
                        bundle=True,
                    )
                if sync:
                    self._server.sync()
        else:
            raise ValueError(f"Unsupported mode '{mode}'.")
        self._allocated = True
        return self

    def load_collection(
        self, data: np.ndarray, mode: str = "file", sr: int = 44100
    ) -> "Buffer":
        """Wrapper method of :func:`Buffer.load_data`"""
        return self.load_data(data, sr=sr, mode=mode)

    def load_asig(self, asig: "pya.Asig", mode: str = "file") -> "Buffer":
        """Create buffer from asig

        Parameters
        ----------
        asig : pya.Asig
            asig to be loaded in buffer
        mode : str, optional
            Insert data via filemode ('file') or n_set OSC commands ('osc'), by default 'file'

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is already allocated.
        """
        if self._allocated:
            raise RuntimeError("Buffer object is already initialized!")
        return self.load_data(asig.sig, sr=asig.sr, mode=mode)

    def use_existing(self, bufnum: int, sr: int = 44100) -> "Buffer":
        """Creates a buffer object from already existing Buffer bufnum.

        Parameters
        ----------
        bufnum : int
            buffer node id
        sr : int
            Sample rate

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is already allocated.
        """
        if self._allocated:
            raise RuntimeError("Buffer object is already initialized!")

        self._alloc_mode = BufferAllocationMode.EXISTING
        self._sr = sr
        self._bufnum = bufnum
        self._allocated = True
        info = self.query()
        self._samples = info.num_frames
        self._channels = info.num_channels
        return self

    def copy_existing(self, buffer: "Buffer") -> "Buffer":
        """Duplicate an existing buffer

        Parameters
        ----------
        buffer : Buffer object
            Buffer which should be duplicated

        Returns
        -------
        self : Buffer
            the newly created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is already allocated.
        """
        if self._allocated:
            raise RuntimeError("Buffer object is already initialized!")
        if not buffer.allocated:
            raise RuntimeError("Other Buffer object is not initialized!")

        # If both buffers use the same server -> copy buffer directly in the server
        if self._server is buffer._server:
            self.alloc(buffer.samples, buffer.sr, buffer.channels)
            self.gen_copy(buffer, 0, 0, -1)
        else:
            # both sc instances must have the same file server
            self._sr = buffer.sr
            tempfile = NamedTemporaryFile(delete=False)
            tempfile.close()
            try:
                buffer.write(tempfile.name)
                self.read(tempfile.name)
            finally:
                if os.path.exists(tempfile.name):
                    os.remove(tempfile.name)
        self._alloc_mode = BufferAllocationMode.COPY
        return self

    # Section: Buffer modification methods
    def fill(self, start: int = 0, count: int = 0, value: float = 0) -> "Buffer":
        """Fill range of samples with value(s).

        Parameters
        ----------
        start : int or list
            int : sample starting index
            list : n*[start, count, value] list
        count : int
            number of samples to fill
        value : float
            value

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        # TODO implement this correctly
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")

        values = [start, count, value] if not isinstance(start, list) else start
        self._server.msg(BufferCommand.FILL, [self._bufnum] + values, bundle=True)
        return self

    def gen(self, command: str, args: List[Any]) -> "Buffer":
        """Call a command to fill a buffer.
        If you know, what you do -> you can use this method.

        See Also
        --------
        gen_sine1, gen_sine2, gen_cheby, gen_cheby, gen_copy

        Parameters
        ----------
        command : str
            What fill command to use.
        args : List[Any]
            Arguments for command

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")
        self._server.msg(BufferCommand.GEN, [self._bufnum, command] + args, bundle=True)
        return self

    def zero(self) -> "Buffer":
        """Set buffer data to zero.

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")
        self._server.msg(BufferCommand.ZERO, [self._bufnum], bundle=True)
        return self

    def gen_sine1(
        self,
        amplitudes: List[float],
        normalize: bool = False,
        wavetable: bool = False,
        clear: bool = False,
    ) -> "Buffer":
        """Fill the buffer with sine waves & given amplitude

        Parameters
        ----------
        amplitudes : list
            The first float value specifies the amplitude of the first partial,
            the second float value specifies the amplitude of the second
            partial, and so on.
        normalize : bool
            Normalize peak amplitude of wave to 1.0.
        wavetable : bool
            If set, then the buffer is written in wavetable format so that it
            can be read by interpolating oscillators.
        clear : bool
            If set then the buffer is cleared before new partials are written
            into it. Otherwise the new partials are summed with the existing
            contents of the buffer

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        return self.gen(
            "sine1", [self._gen_flags(normalize, wavetable, clear), amplitudes]
        )

    def gen_sine2(
        self,
        freq_amps: List[float],
        normalize: bool = False,
        wavetable: bool = False,
        clear: bool = False,
    ) -> "Buffer":
        """Fill the buffer with sine waves
        given list of [frequency, amplitude] lists

        Parameters
        ----------
        freq_amps : list
            Similar to sine1 except that each partial frequency is specified
            explicitly instead of being an integer multiple of the fundamental.
            Non-integer partial frequencies are possible.
        normalize : bool
            If set, normalize peak amplitude of wave to 1.0.
        wavetable : bool
            If set, the buffer is written in wavetable format so that it
            can be read by interpolating oscillators.
        clear : bool
            If set, the buffer is cleared before new partials are written
            into it. Otherwise the new partials are summed with the existing
            contents of the buffer.

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        return self.gen(
            "sine2", [self._gen_flags(normalize, wavetable, clear), freq_amps]
        )

    def gen_sine3(
        self,
        freqs_amps_phases: List[float],
        normalize: bool = False,
        wavetable: bool = False,
        clear: bool = False,
    ) -> "Buffer":
        """Fill the buffer with sine waves & given a list of
        [frequency, amplitude, phase] entries.

        Parameters
        ----------
        freqs_amps_phases : list
            Similar to sine2 except that each partial may have a
            nonzero starting phase.
        normalize : bool
            if set, normalize peak amplitude of wave to 1.0.
        wavetable : bool
            If set, the buffer is written in wavetable format
            so that it can be read by interpolating oscillators.
        clear : bool
            If set, the buffer is cleared before new partials are written
            into it. Otherwise the new partials are summed with the existing
            contents of the buffer.

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        return self.gen(
            "sine3", [self._gen_flags(normalize, wavetable, clear), freqs_amps_phases]
        )

    def gen_cheby(
        self,
        amplitudes: List[float],
        normalize: bool = False,
        wavetable: bool = False,
        clear: bool = False,
    ) -> "Buffer":
        """Fills a buffer with a series of chebyshev polynomials, which can be
        defined as cheby(n) = amplitude * cos(n * acos(x))

        Parameters
        ----------
        amplitudes : list
            The first float value specifies the amplitude for n = 1,
            the second float value specifies the amplitude
            for n = 2, and so on
        normalize : bool
            If set, normalize the peak amplitude of the Buffer to 1.0.
        wavetable : bool
            If set, the buffer is written in wavetable format so that it
            can be read by interpolating oscillators.
        clear : bool
            If set the buffer is cleared before new partials are written
            into it. Otherwise the new partials are summed with the existing
            contents of the buffer.

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        return self.gen(
            "cheby", [self._gen_flags(normalize, wavetable, clear), amplitudes]
        )

    def gen_copy(
        self, source: "Buffer", source_pos: int, dest_pos: int, copy_amount: int
    ) -> "Buffer":
        """Copy samples from the source buffer to the destination buffer
        specified in the b_gen command.

        Parameters
        ----------
        source : Buffer
            Source buffer object
        source_pos : int
            sample position in source
        dest_pos : int
            sample position in destination
        copy_amount : int
            number of samples to copy. If the number of samples to copy is
            negative, the maximum number of samples
            possible is copied.

        Returns
        -------
        self : Buffer
            the created Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        return self.gen("copy", [dest_pos, source.bufnum, source_pos, copy_amount])

    # Section: Buffer output methods
    def play(
        self, rate: float = 1, loop: bool = False, pan: float = 0, amp: float = 0.3
    ) -> Synth:
        """Play the Buffer using a Synth

        Parameters
        ----------
        rate : float, optional
            plackback rate, by default 1
        loop : bool, optional
            if True loop the playback, by default False
        pan : int, optional
            pan position, -1 is left, +1 is right, by default 0
        amp : float, optional
            amplitude, by default 0.3

        Returns
        -------
        Synth
            Synth to control playback.

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")

        if self._synth_def is None:
            playbuf_def = """
            { |out=0, bufnum=^bufnum, rate=^rate, loop=^loop, pan=^pan, amp=^amp |
                    var sig = PlayBuf.ar(^num_channels, bufnum,
                        rate*BufRateScale.kr(bufnum),
                        loop: loop,
                        doneAction: Done.freeSelf);
                    Out.ar(out, Pan2.ar(sig, pan, amp))
            }"""
            self._synth_def = SynthDef(
                name=f"sc3nb_playbuf_{self.bufnum}", definition=playbuf_def
            )
            synth_name = self._synth_def.add(
                pyvars={
                    "num_channels": self.channels,
                    "bufnum": self.bufnum,
                    "rate": rate,
                    "loop": 1 if loop else 0,
                    "pan": pan,
                    "amp": amp,
                }
            )
            self._synth = Synth(name=synth_name, server=self._server)
        else:
            self._synth.new(
                {"rate": rate, "loop": 1 if loop else 0, "pan": pan, "amp": amp}
            )
        return self._synth

    def write(
        self,
        path: str,
        header: str = "wav",
        sample: str = "float",
        num_frames: int = -1,
        starting_frame: int = 0,
        leave_open: bool = False,
    ) -> "Buffer":
        """Write buffer data to a sound file

        Parameters
        ----------
        path : string
            path name of a sound file.
        header : string
            header format. Header format is one of:
            "aiff", "next", "wav", "ircam"", "raw"
        sample : string
            sample format. Sample format is one of:
            "int8", "int16", "int24", "int32",
            "float", "double", "mulaw", "alaw"
        num_frames : int
            number of frames to write.
            -1 means all frames.
        starting_frame : int
            starting frame in buffer
        leave_open : boolean
            Whether you want the buffer file left open.
            For use with DiskOut you will want this to be true.
            The file is created, but no frames are written until the DiskOut UGen does so.
            The default is false which is the correct value for all other cases.

        Returns
        -------
        self : Buffer
            the Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")
        leave_open_val = 1 if leave_open else 0
        path = str(Path(path).resolve())
        self._server.msg(
            BufferCommand.WRITE,
            [
                self._bufnum,
                path,
                header,
                sample,
                num_frames,
                starting_frame,
                leave_open_val,
            ],
            bundle=True,
        )
        return self

    def close(self) -> "Buffer":
        """Close soundfile after using a Buffer with DiskOut

        Returns
        -------
        self : Buffer
            the Buffer object

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")
        self._server.msg(BufferCommand.CLOSE, [self._bufnum], bundle=True)
        return self

    def to_array(self) -> np.ndarray:
        """Return the buffer data as an array representation.

        Returns
        -------
        np.ndarray:
            Values of the buffer

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")
        data = []
        blocksize = 1000  # array size compatible with OSC packet size
        i = 0
        num_samples = self._samples * self._channels
        while i < num_samples:
            bs = blocksize if i + blocksize < num_samples else num_samples - i
            tmp = self._server.msg(
                BufferCommand.GETN, [self._bufnum, i, bs], bundle=False
            )
            data += list(tmp)[3:]  # skip first 3 els [bufnum, startidx, size]
            i += bs
        data = np.array(data).reshape((-1, self._channels))
        return data

    # Section: Buffer information methods
    def query(self) -> BufferInfo:
        """Get buffer info.

        Returns
        -------
        Tuple:
            (buffer number, number of frames, number of channels, sampling rate)

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")
        return BufferInfo._make(
            self._server.msg(BufferCommand.QUERY, [self._bufnum], bundle=False)
        )

    def __repr__(self) -> str:
        if self.samples is None or self.sr is None:
            duration = 0
        else:
            duration = self.samples / self.sr
        return (
            f"<Buffer({self.bufnum}) on {self._server.addr}:"
            + f" {self.channels} x {self.samples} @ {self.sr} Hz = {duration:.3f}s"
            + f""" {["not loaded", "allocated"][self.allocated]}"""
            + f" using mode '{self._alloc_mode}'>"
        )

    # Section: Methods to delete / free Buffers
    def free(self) -> None:
        """Free buffer data.

        Raises
        ------
        RuntimeError
            If the Buffer is not allocated yet.
        """
        if not self._allocated:
            raise RuntimeError("Buffer object is not initialized!")
        if (
            self._alloc_mode != BufferAllocationMode.EXISTING
            and not self._bufnum_set_manually
        ):
            self._server.buffer_ids.free([self._bufnum])
        self._server.msg(BufferCommand.FREE, [self._bufnum], bundle=True)
        self._allocated = False
        self._alloc_mode = BufferAllocationMode.NONE

    # Section: Properties
    @property
    def bufnum(self) -> Optional[int]:
        """Buffer number which serves as ID in SuperCollider

        Returns
        -------
        int
            bufnum
        """
        return self._bufnum

    @property
    def allocated(self) -> bool:
        """Whether this Buffer is allocated by
        any of the initialization methods.

        Returns
        -------
        bool
            True if allocated
        """
        return self._allocated

    @property
    def alloc_mode(self) -> BufferAllocationMode:
        """Mode of Buffer allocation.

        One of ['file', 'alloc', 'data', 'existing', 'copy']
        according to previously used generator.
        Defaults to None if not allocated.

        Returns
        -------
        str
            allocation mode
        """
        return self._alloc_mode

    @property
    def path(self) -> Optional[Path]:
        """File path that was provided to read.

        Returns
        -------
        pathlib.Path
            buffer file path
        """
        return self._path

    @property
    def channels(self) -> Optional[int]:
        """Number of channels in the Buffer.

        Returns
        -------
        int
            channel number
        """
        return self._channels

    @property
    def samples(self) -> Optional[int]:
        """Number of samples in the buffer.

        Returns
        -------
        int
            sample number
        """
        return self._samples

    @property
    def sr(self) -> Optional[int]:
        """Sampling rate of the Buffer.

        Returns
        -------
        int
            sampling rate
        """
        return self._sr

    @property
    def duration(self) -> Optional[float]:
        """Duration of the Buffer in seconds.

        Returns
        -------
        float
            duration in seconds
        """
        if self._samples is not None and self._sr is not None:
            return self._samples / self._sr
        else:
            return None

    @property
    def server(self) -> "SCServer":
        """The server where this Buffer is placed.

        Returns
        -------
        SCServer
            The server where this Buffer is placed.
        """
        return self._server

    # Section: Private utils
    def _gen_flags(self, a_normalize=False, a_wavetable=False, a_clear=False) -> int:
        """Generate Wave Fill Commands flags from booleans
        according to the SuperCollider Server Command Reference.

        Parameters
        ----------
        a_normalize : bool, optional
            Normalize peak amplitude of wave to 1.0, by default False
        a_wavetable : bool, optional
            If set, then the buffer is written in wavetable
            format so that it can be read by interpolating
            oscillators, by default False
        a_clear : bool, optional
            If set then the buffer is cleared before new partials are written
            into it. Otherwise the new partials are summed with the existing
            contents of the buffer, by default False

        Returns
        -------
        int
            Wave Fill Commands flags
        """
        normalize = 1 if a_normalize is True else 0
        wavetable = 2 if a_wavetable is True else 0
        clear = 4 if a_clear is True else 0
        return normalize + wavetable + clear



from unittest import TestCase

from sc3nb.sc_objects.node import Group
from sc3nb.sc_objects.server import SCServer, ServerOptions


class ServerTest(TestCase):
    def setUp(self) -> None:
        self.port = 57777
        options = ServerOptions(udp_port=self.port)
        self.server = SCServer(options=options)
        self.server.boot(with_blip=False)

    def tearDown(self) -> None:
        self.server.quit()

    def test_connection_info(self):
        _, receivers = self.server.connection_info()
        ip_address, port = next(iter(receivers.keys()))
        self.assertEqual(len(receivers), 1)
        self.assertEqual(ip_address, "127.0.0.1")
        self.assertEqual(port, self.port)

    def test_is_local(self):
        self.assertTrue(self.server.is_local)

    def test_sync(self):
        self.assertTrue(self.server.sync(timeout=10))

    def test_node_tree(self):
        root_group = self.server.query_tree()
        self.assertEqual(len(root_group.children), self.server.max_logins)
        self.assertIn(self.server.default_group, root_group.children)
        group = Group(server=self.server)
        self.server.sync()
        root_group = self.server.query_tree()
        self.assertIn(group, self.server.default_group.children)

    def test_multiclient(self):
        options = ServerOptions(udp_port=self.port)
        self.other_server = SCServer(options=options)
        self.other_server._max_warning_freq = float("-inf")
        with self.assertRaisesRegex(
            ValueError,
            f"The specified UDP port {self.server.options.udp_port} is already used",
        ):
            self.other_server.boot(timeout=3)
        self.other_server.remote(*self.server.addr)
        self.assertEqual(self.server.addr, self.other_server.addr)
        # check if they share nodes
        group = Group(server=self.other_server)
        root_group = self.server.query_tree()
        matches = [
            child
            for child in root_group.children
            if child.nodeid == self.other_server.default_group.nodeid
        ]
        self.assertEqual(len(matches), 1)
        other_default_group: Group = matches[0]
        self.assertIn(
            group.nodeid, [child.nodeid for child in other_default_group.children]
        )



#!/usr/bin/env python

import argparse
import copy
import errno
import glob
import os
import shutil
import stat
import sys
from pathlib import Path
from sys import platform
from typing import Optional, Sequence

from bs4 import BeautifulSoup
from git import Repo


def main():
    repo = Repo(search_parent_directories=True)
    git_root = repo.working_dir
    if git_root is None:
        raise RuntimeError("No git working dir found.")
    os.chdir(git_root)

    parser = argparse.ArgumentParser(description="Generates Sphinx documentation")
    parser.add_argument("--github", action="store_true", help="build gh-pages")
    parser.add_argument(
        "--commit", action="store_true", help="create a commit (only when --github)"
    )
    parser.add_argument(
        "--show", action="store_true", help="open in browser after build"
    )
    parser.add_argument(
        "--branches",
        nargs="*",
        default=["master", "develop"],
        help="limit build to these branches (only when --github)",
    )
    parser.add_argument(
        "--tags",
        nargs="*",
        default=None,
        help="limit build to these tags (only when --github)",
    )
    parser.add_argument(
        "--input",
        default=f"{git_root}/docs/source/",
        help="input folder (ignored for gh-pages)",
    )
    parser.add_argument(
        "--output", default=f"{git_root}/build/docs/", help="output folder"
    )
    parser.add_argument(
        "--clean", action="store_true", help="removes out folder before building doc"
    )
    args = parser.parse_args()

    # preparing output folder
    if args.clean and os.path.exists(args.output):
        print(f"Cleaning {args.output}")
        rm_dir(args.output)
    os.makedirs(args.output, exist_ok=True)

    if not args.github:
        target = "html"
        docs_root = build_doc(
            source=Path(args.input), out=Path(args.output), target=target
        )
    else:
        docs_root = generate_gh_pages(
            repo=repo,
            out_folder=Path(args.output),
            branches=args.branches,
            tags=args.tags,
            commit=args.commit,
            clean=args.clean,
        )
    if args.show:
        url = docs_root
        if platform == "darwin":
            os.system(f"open {url}")
        elif platform == "win32":
            os.startfile(url)
        else:
            raise NotImplementedError("'show' is not available for your platform")

    # if args.show:
    #    _run_webserver(os.path.normpath(repo))


def fix_all_links(path, versions):
    def change_dd_tag(tag, rst_version):
        new_tag = copy.copy(tag)
        splits = new_tag.a["href"].split("/")
        splits[-2] = rst_version
        new_tag.a["href"] = "/".join(splits)
        new_tag.a.string = rst_version
        return new_tag

    def _fix_links(path, versions):
        with open(path, "r", encoding="utf8", errors="ignore") as file:
            content = file.read()
            soup = BeautifulSoup(content, features="html.parser")
        divs = soup.find_all("div", class_="rst-versions")
        if len(divs) != 1:
            print(f"Matched {len(divs)} divs in file: {path}")
        else:
            div = divs[0]
            if "/repo/stable/" in Path(path).as_posix():
                current_version = "stable"
            else:
                current_version = (
                    div.find("span", class_="rst-current-version")
                    .contents[2]
                    .strip()
                    .split()[1]
                )
            dl = div.find("dl")
            dt = dl.dt.extract()
            dd = dl.dd.extract()
            dds = [change_dd_tag(dd, version) for version in versions]
            dl.clear()
            dl.insert(0, dt)
            dl.extend(dds)
            for dd in dds:
                if dd.text == current_version:
                    strong = soup.new_tag("strong")
                    dd.wrap(strong)
            # write file
            with open(path, "w", encoding="utf8") as file:
                file.write(str(soup))

    files = glob.glob(f"{path}/**/*.html", recursive=True)
    for file in files:
        print(f"Fixing links in {file}")
        _fix_links(file, versions)


def _handle_remove_readonly(func, path, exc):
    excvalue = exc[1]
    if (
        func not in (os.rmdir, os.remove, os.unlink, os.scandir)
        or excvalue.errno != errno.EACCES
    ):
        raise RuntimeError(f"Failed to remove {path}")

    os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
    func(path)


def rm_dir(path):
    shutil.rmtree(path, ignore_errors=False, onerror=_handle_remove_readonly)


def build_doc(
    source: Path, out: Path, target: str, additional_options: Optional[str] = ""
):
    print(f"Starting sphinx-build for target {target} {additional_options} ...")
    # prevents __pycache__ files which we don't need when code runs only once.
    docs_out = out / target
    sys.dont_write_bytecode = True
    exec_str = f"sphinx-build -b {target} {additional_options} {source} {docs_out}"
    print(exec_str)
    res = os.system(exec_str)
    if res != 0:
        raise RuntimeError(
            f"Could not generate documentation for {target} {additional_options}. Build returned status code {res}!"
        )
    print(f"Done generating documentation for {target} {additional_options}.")
    return docs_out


def generate_gh_pages(
    repo: Repo,
    out_folder: Path,
    branches: Sequence[str] = ("master", "develop"),
    tags: Sequence[str] = (),
    commit: bool = False,
    clean: bool = False,
):
    print("Generate github pages")
    gh_pages_root = out_folder / "gh-pages"
    build_folder = gh_pages_root / "out"
    if build_folder.exists():
        rm_dir(build_folder)
    os.makedirs(build_folder)

    tmp_repo_path = gh_pages_root / "repo"
    if tmp_repo_path.exists() and not clean:
        tmp_repo = Repo(tmp_repo_path)
        tmp_repo.git.checkout("--guess", "gh-pages", "--")
        print(f"Pulling repo...")
        tmp_repo.git.pull("-a")
    else:
        print(f"Cloning repo...")
        tmp_repo = Repo.clone_from(repo.remote().url, tmp_repo_path)

    print("Checking repo...")
    tmp_repo.git.clean("--force", "-d", "-x")
    tmp_repo.git.checkout("--guess", "gh-pages", "--")

    documented = [
        p.name
        for p in tmp_repo_path.iterdir()
        if p.is_dir() and p.parent == tmp_repo_path and not p.name.startswith(".")
    ]

    if not tags:

        def check_version(tag, min_version):
            return tuple(map(int, tag[1:].split("."))) >= min_version

        tags = [tag.name for tag in tmp_repo.tags if check_version(tag.name, (1, 0, 1))]

    # first, check which documentation to generate
    def has_docs(name):
        previous_branch = tmp_repo.active_branch
        tmp_repo.git.checkout("--guess", name, "--")
        docs_found = (tmp_repo_path / "docs").exists()
        tmp_repo.git.checkout("--guess", previous_branch, "--")
        return docs_found

    branches = [b for b in branches if has_docs(b)]
    tags = [t for t in tags if has_docs("tags/" + t) and t[1:] not in documented]
    print(
        f"Will generate documentation for:\n  already documented: {documented}\n  Branches: {branches}\n  Tags: {tags}"
    )
    stable = tags[-1] if tags else None
    print(f"new stable version {stable}")
    # fix order of dropdown elements (most recent tag first, then branches and older tags)
    doclist = branches + tags
    print(f"doclist={doclist}")
    # generate documentation; all versions have to be known for the dropdown menu
    for d in doclist:
        print(f"Generating documentation for {d} ...")
        target = d if d in branches else "tags/" + d
        tmp_repo.git.checkout("--guess", target, "--")
        tmp_repo.git.clean("--force", "-d", "-x")
        build_doc(
            source=tmp_repo_path / "docs/source",
            out=build_folder / d,
            target="html",
        )

    # create index html to forward to last tagged version
    with open(f"{build_folder}/index.html", "w") as fp:
        fp.write(
            f"""<!DOCTYPE html>
<html>
  <head>
    <meta http-equiv="refresh" content="0; url=stable/">
  </head>
</html>
"""
        )

    # prepare gh-pages
    tmp_repo.git.checkout("--guess", "gh-pages", "--")
    tmp_repo.git.clean("--force", "-d", "-x")

    print(f"Merging documentation in {tmp_repo_path}...")
    for item in build_folder.iterdir():
        s = build_folder / item.name
        d = tmp_repo_path / item.name
        print(f" - {item.name}: {s} -> {d}")
        if s.is_dir():
            s = s / "html"
            if d.exists():
                shutil.rmtree(d, ignore_errors=True)
            shutil.copytree(s, d, copy_function=shutil.copy)
        else:
            shutil.copy(s, d)
        if stable is not None and item == stable:
            print(f"Copying {s} as new stable version")
            stable_path = tmp_repo_path / "stable"
            shutil.rmtree(stable_path, ignore_errors=True)
            shutil.copytree(s, stable_path, copy_function=shutil.copy)

    print(f"Documentation has been merged in {tmp_repo.working_dir}")
    # refresh the documented list
    documented = [
        p.name
        for p in tmp_repo_path.iterdir()
        if p.is_dir() and p.parent == tmp_repo_path and not p.name.startswith(".")
    ]
    print(f"Fixing links to other versions {documented}")
    fix_all_links(tmp_repo.working_dir, versions=documented)
    print("Current git status:")
    status = tmp_repo.git.status("--short")
    print(status)
    if commit and status:
        print("Creating commit")
        with tmp_repo.config_writer() as writer:
            writer.set_value(
                "user", "name", repo.config_reader().get_value("user", "name")
            )
            writer.set_value(
                "user", "email", repo.config_reader().get_value("user", "email")
            )
        tmp_repo.git.add(".")
        tmp_repo.git.commit("-m Update docs")
        print(f"Control the commit in {tmp_repo.working_dir}")

    return gh_pages_root


if __name__ == "__main__":
    main()



"""Module with utlilty functions - especially for handling code snippets"""
import inspect
import re
import socket
from typing import Any

import numpy as np


def is_socket_used(addr=("127.0.0.1", 57110)):
    a_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        result_of_check = a_socket.bind(addr)
    except OSError as error:
        return True
    else:
        return False
    finally:
        a_socket.close()


def remove_comments(code: str) -> str:
    """Removes all c-style comments from code.

    This removes `//single-line` or `/* multi-line */`  comments.

    Parameters
    ----------
    code : str
        Code where comments should be removed.

    Returns
    -------
    str
        code string without comments
    """
    # function code by Onur Yildirim (https://stackoverflow.com/a/18381470)
    # first group captures quoted strings (double or single)
    # second group captures comments (//single-line or /* multi-line */)
    # alternative cares for escaped quotes
    # pattern = r"(\".*?(?<!\\)\"|\'.*?(?<!\\)\')|(/\*.*?\*/|//[^\r\n]*$)"
    pattern = r"(\".*?\"|\'.*?\')|(/\*.*?\*/|//[^\r\n]*$)"
    regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

    def _replacer(match):
        # if the 2nd group (capturing comments) is not None,
        # it means we have captured a non-quoted (real) comment string.
        if match.group(2) is not None:
            return ""  # so we will return empty to remove the comment
        else:  # otherwise, we will return the 1st group
            return match.group(1)  # captured quoted-string

    return regex.sub(_replacer, code)


def parse_pyvars(code: str, frame_nr: int = 2):
    """Looks through call stack and finds values of variables.

    Parameters
    ----------
    code : str
        SuperCollider command to be parsed
    frame_nr : int, optional
        on which frame to start, by default 2 (grandparent frame)

    Returns
    -------
    dict
        {variable_name: variable_value}

    Raises
    ------
    NameError
        If the variable value could not be found.
    """
    matches = re.findall(r"\s*\^[A-Za-z_]\w*\s*", code)

    pyvars = {match.split("^")[1].strip(): None for match in matches}
    missing_vars = list(pyvars.keys())

    stack = inspect.stack()
    frame = None
    try:
        while missing_vars and frame_nr < len(stack):
            frame = stack[frame_nr][0]
            for pyvar in pyvars:
                if pyvar not in missing_vars:
                    continue
                # check for variable in local variables
                if pyvar in frame.f_locals:
                    pyvars[pyvar] = frame.f_locals[pyvar]
                    missing_vars.remove(pyvar)
                # check for variable in global variables
                elif pyvar in frame.f_globals:
                    pyvars[pyvar] = frame.f_globals[pyvar]
                    missing_vars.remove(pyvar)
            frame_nr += 1
    finally:
        del frame
        del stack
    if missing_vars:
        raise NameError("name(s) {} not defined".format(missing_vars))
    return pyvars


def replace_vars(code: str, pyvars: dict) -> str:
    """Replaces python variables with SuperCollider literals in code.

    This replaces the pyvars preceded with ^ in the code with a SC literal.
    The conversion is done with convert_to_sc.

    Parameters
    ----------
    code : str
        SuperCollider Code with python injections.
    pyvars : dict
        Dict with variable names and values.

    Returns
    -------
    str
        Code with injected variables.
    """
    for pyvar, value in pyvars.items():
        pyvar = "^" + pyvar
        value = convert_to_sc(value)
        code = code.replace(pyvar, value)
    return code


def convert_to_sc(obj: Any) -> str:
    """Converts python objects to SuperCollider code literals.

    This supports currently:

    * numpy.ndarray -> SC Array representation
    * complex type -> SC Complex
    * strings -> if starting with sc3: it will be used as SC code
                 if it starts with a \\ (single escaped backward slash) it will be used as symbol
                 else it will be inserted as string

    For unsupported types the __repr__ will be used.

    Parameters
    ----------
    obj : Any
        object that should be converted to a SuperCollider code literal.

    Returns
    -------
    str
        SuperCollider Code literal
    """
    if isinstance(obj, np.ndarray):
        return obj.tolist().__repr__()
    if isinstance(obj, complex):
        return "Complex({0}, {1})".format(obj.real, obj.imag)
    if isinstance(obj, str):
        if obj.startswith("sc3:"):  # start sequence for sc3-code
            return "{}".format(obj[4:])
        if obj.startswith(r"\\") and not obj.startswith(r"\\\\"):
            return "'{}'".format(obj[1:])  # 'x' will be interpreted as symbol
        if obj.startswith(r"\\\\"):
            obj = obj[1:]
        return '"{}"'.format(obj)  # "x" will be interpreted as string
    # further type conversion can be added in the future
    return obj.__repr__()



"""Server Volume controls."""


import logging
import warnings
from typing import TYPE_CHECKING, Optional

from pyamapping import clip, db_to_amp

from sc3nb.sc_objects.node import AddAction, Synth
from sc3nb.sc_objects.synthdef import SynthDef

if TYPE_CHECKING:
    from sc3nb.sc_objects.server import SCServer


_LOGGER = logging.getLogger(__name__)


class Volume:
    """Server volume controls"""

    def __init__(self, server: "SCServer", min_: int = -90, max_: int = 6) -> None:
        self._server = server
        self._server.add_init_hook(self.send__volume_synthdef)
        self._server.add_init_hook(self.update_volume_synth)

        self.min = min_
        self.max = max_

        self._muted = False
        self._volume = 0.0
        self._lag = 0.1

        self._synth_name: Optional[str] = None

        self._synth: Optional[Synth] = None

    @property
    def muted(self):
        """True if muted."""
        return self._muted

    @muted.setter
    def muted(self, muted: bool):
        if muted:
            self.mute()
        else:
            self.unmute()

    @property
    def volume(self):
        """Volume in dB."""
        return self._volume

    @volume.setter
    def volume(self, volume):
        self._volume = clip(volume, self.min, self.max)
        self.update_volume_synth()

    def mute(self) -> None:
        """Mute audio"""
        self._muted = True
        self.update_volume_synth()

    def unmute(self) -> None:
        """Unmute audio"""
        self._muted = False
        self.update_volume_synth()

    def update_volume_synth(self) -> None:
        """Update volume Synth"""
        amp = 0.0 if self._muted else db_to_amp(self._volume)
        active = amp != 1.0
        if active:
            if self._server.is_running:
                if self._synth is None:
                    if self._synth_name is None:
                        warnings.warn(
                            "Cannot set volume. Volume SynthDef unknown. Is the default sclang running?"
                        )
                        return
                    controls = {
                        "volumeAmp": amp,
                        "volumeLag": self._lag,
                        "bus": self._server.output_bus.idxs[0],
                    }
                    self._synth = Synth(
                        self._synth_name,
                        add_action=AddAction.AFTER,
                        target=self._server.default_group,
                        controls=controls,
                        server=self._server,
                    )
                else:
                    self._synth.set("volumeAmp", amp)
        else:
            if self._synth is not None:
                self._synth.release()
                self._synth = None

    def send__volume_synthdef(self):
        """Send Volume SynthDef"""
        if self._server.is_running:
            num_channels = self._server.output_bus.num_channels
            synth_def = SynthDef(
                f"sc3nb_volumeAmpControl{num_channels}",
                r"""{ | volumeAmp = 1, volumeLag = 0.1, gate=1, bus |
                    XOut.ar(bus,
                        Linen.kr(gate, releaseTime: 0.05, doneAction:2),
                        In.ar(bus, ^num_channels) * Lag.kr(volumeAmp, volumeLag)
                    );
                }""",
            )
            try:
                self._server.lookup_receiver("sclang")
            except KeyError:
                _LOGGER.info(
                    "Volume SynthDef cannot be send. No sclang receiver known."
                )
            else:
                self._synth_name = synth_def.add(server=self._server)
                assert self._synth_name is not None, "Synth name is None"



"""This module adds the Jupyter specialties such as Magics and Keyboard Shortcuts"""
import re
import sys
import warnings
from typing import Any, Dict

from IPython.core.magic import Magics, line_cell_magic, magics_class

import sc3nb


def load_ipython_extension(ipython) -> None:
    """Function that is called when Jupyter loads this as extension (%load_ext sc3nb)

    Parameters
    ----------
    ipython : IPython
        IPython object
    """
    ipython.register_magics(SC3Magics)
    add_shortcut(ipython)


def add_shortcut(ipython, shortcut: str = None) -> None:
    """Add the server 'free all' shortcut.

    Parameters
    ----------
    ipython : IPython
        IPython object
    shortcut : str, optional
        shortcut for 'free all', by default it is "cmd-." or "Ctrl-."
    """
    try:
        if shortcut is None:
            if sys.platform == "darwin":
                shortcut = "cmd-."
            elif sys.platform in ["linux", "linux2", "win32"]:
                shortcut = "Ctrl-."
            else:
                warnings.warn(f"Unable to add shortcut for platform '{sys.platform}'")
        if shortcut is not None:
            ipython.run_cell_magic(
                "javascript",
                "",
                f"""if (typeof Jupyter !== 'undefined') {{
                        Jupyter.keyboard_manager.command_shortcuts.add_shortcut(
                        \'{shortcut}\', {{
                        help : \'Free all nodes on SC server\',
                        help_index : \'zz\',
                        handler : function (event) {{
                            IPython.notebook.kernel.execute(
                                "import sc3nb; sc3nb.SC.get_default().server.free_all(root=True)"
                            )
                            return true;}}
                        }});
                    }}""",
            )
    except AttributeError:
        pass


@magics_class
class SC3Magics(Magics):
    """IPython magics for SC class"""

    @line_cell_magic
    def sc(self, line="", cell=None):
        """Execute SuperCollider code via magic

        Parameters
        ----------
        line : str, optional
            Line of SuperCollider code, by default ''
        cell : str, optional
            Cell of SuperCollider code , by default None

        Returns
        -------
        Unknown
            cmd result
        """
        cmdstr = line if cell is None else cell
        pyvars = self._parse_pyvars(cmdstr)
        return sc3nb.SC.get_default().lang.cmd(cmdstr, pyvars=pyvars)

    @line_cell_magic
    def scv(self, line="", cell=None):
        """Execute SuperCollider code with verbose output

        Parameters
        ----------
        line : str, optional
            Line of SuperCollider code, by default ''
        cell : str, optional
            Cell of SuperCollider code , by default None

        Returns
        -------
        Unknown
            cmd result
        """
        cmdstr = line if cell is None else cell
        pyvars = self._parse_pyvars(cmdstr)
        return sc3nb.SC.get_default().lang.cmdv(cmdstr, pyvars=pyvars)

    @line_cell_magic
    def scs(self, line="", cell=None):
        """Execute SuperCollider code silently (verbose==False)

        Parameters
        ----------
        line : str, optional
            Line of SuperCollider code, by default ''
        cell : str, optional
            Cell of SuperCollider code , by default None

        Returns
        -------
        Unknown
            cmd result
        """
        cmdstr = line if cell is None else cell
        pyvars = self._parse_pyvars(cmdstr)
        return sc3nb.SC.get_default().lang.cmds(cmdstr, pyvars=pyvars)

    @line_cell_magic
    def scg(self, line="", cell=None):
        """Execute SuperCollider code returning output

        Parameters
        ----------
        line : str, optional
            Line of SuperCollider code, by default ''
        cell : str, optional
            Cell of SuperCollider code , by default None

        Returns
        -------
        Unknown
            cmd result
            Output from SuperCollider code, not
            all SC types supported, see
            pythonosc.osc_message.Message for list
            of supported types
        """
        cmdstr = line if cell is None else cell
        pyvars = self._parse_pyvars(cmdstr)
        return sc3nb.SC.get_default().lang.cmdg(cmdstr, pyvars=pyvars)

    @line_cell_magic
    def scgv(self, line="", cell=None):
        """Execute SuperCollider code returning output

        Parameters
        ----------
        line : str, optional
            Line of SuperCollider code, by default ''
        cell : str, optional
            Cell of SuperCollider code , by default None

        Returns
        -------
        Unknown
            cmd result
            Output from SuperCollider code, not
            all SC types supported, see
            pythonosc.osc_message.Message for list
            of supported types
        """
        cmdstr = line if cell is None else cell
        pyvars = self._parse_pyvars(cmdstr)
        return sc3nb.SC.get_default().lang.cmdg(cmdstr, pyvars=pyvars, verbose=True)

    @line_cell_magic
    def scgs(self, line="", cell=None):
        """Execute SuperCollider code returning output

        Parameters
        ----------
        line : str, optional
            Line of SuperCollider code, by default ''
        cell : str, optional
            Cell of SuperCollider code , by default None

        Returns
        -------
        Unknown
            cmd result
            Output from SuperCollider code, not
            all SC types supported, see
            pythonosc.osc_message.Message for list
            of supported types
        """
        cmdstr = line if cell is None else cell
        pyvars = self._parse_pyvars(cmdstr)
        return sc3nb.SC.get_default().lang.cmdg(cmdstr, pyvars=pyvars, verbose=False)

    def _parse_pyvars(self, code: str) -> Dict[str, Any]:
        """Parses SuperCollider code for python variables and their values

        Parameters
        ----------
        code : str
            SuperCollider code snippet

        Returns
        -------
        Dict[str, Any]
            Dict with variable names and their values.

        Raises
        ------
        NameError
            If pyvar injection value can't be found.
        """
        user_ns = self.shell.user_ns

        matches = re.findall(r"\s*\^[A-Za-z_]\w*\s*", code)
        pyvars = {match.split("^")[1].strip(): None for match in matches}
        for pyvar in pyvars:
            if pyvar in user_ns:
                pyvars[pyvar] = user_ns[pyvar]
            else:
                raise NameError("name '{}' is not defined".format(pyvar))

        return pyvars



"""Module to for using SuperCollider SynthDefs and Synths in Python"""

import re
import sys
import warnings
from enum import Enum, unique
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

if sys.version_info < (3, 9):
    # `importlib.resources` backported to PY<37 as `importlib_resources`.
    import importlib_resources as libresources
else:
    # only PY>=39 `importlib.resources` offers .files.
    import importlib.resources as libresources

import sc3nb
import sc3nb.resources
from sc3nb.util import parse_pyvars, replace_vars

if TYPE_CHECKING:
    from sc3nb.sc import SC
    from sc3nb.sc_objects.server import SCServer
    from sc3nb.sclang import SCLang, SynthArgument


@unique
class SynthDefinitionCommand(str, Enum):
    """OSC Commands for Synth Definitions"""

    RECV = "/d_recv"
    LOAD = "/d_load"
    LOAD_DIR = "/d_loadDir"
    FREE = "/d_free"


class SynthDef:
    """Wrapper for SuperCollider SynthDef"""

    synth_descs = {}
    synth_defs = {}

    @classmethod
    def get_description(
        cls, name: str, lang: Optional["SCLang"] = None
    ) -> Optional[Dict[str, "SynthArgument"]]:
        """Get Synth description

        Parameters
        ----------
        name : str
            name of SynthDef

        Returns
        -------
        Dict
            dict with SynthArguments
        """
        if name in cls.synth_descs:
            return cls.synth_descs[name]
        synth_desc = None
        if lang is None:
            try:
                lang = sc3nb.SC.get_default().lang
            except RuntimeError:
                warnings.warn(f"SynthDesc '{name}' is unknown. sclang is not running.")
                return synth_desc
        try:
            synth_desc = lang.get_synth_description(name)
        except ValueError:
            warnings.warn(
                f"SynthDesc '{name}' is unknown. sclang does not know this SynthDef"
            )

        if synth_desc is not None:
            cls.synth_descs[name] = synth_desc
        return synth_desc

    @classmethod
    def send(
        cls,
        synthdef_bytes: bytes,
        server: Optional["SCServer"] = None,
    ):
        """Send a SynthDef as bytes.

        Parameters
        ----------
        synthdef_bytes : bytes
            SynthDef bytes
        wait : bool
            If True wait for server reply.
        server : SCServer, optional
            Server instance that gets the SynthDefs,
            by default use the SC default server
        """
        if server is None:
            server = sc3nb.SC.get_default().server
        server.msg(
            SynthDefinitionCommand.RECV,
            synthdef_bytes,
            await_reply=True,
            bundle=True,
        )

    @classmethod
    def load(
        cls,
        synthdef_path: str,
        server: Optional["SCServer"] = None,
    ):
        """Load SynthDef file at path.

        Parameters
        ----------
        synthdef_path : str
            Path with the SynthDefs
        server : SCServer, optional
            Server that gets the SynthDefs,
            by default use the SC default server
        """
        if server is None:
            server = sc3nb.SC.get_default().server
        server.msg(
            SynthDefinitionCommand.LOAD,
            synthdef_path,
            await_reply=True,
            bundle=True,
        )

    @classmethod
    def load_dir(
        cls,
        synthdef_dir: Optional[str] = None,
        completion_msg: Optional[bytes] = None,
        server: Optional["SCServer"] = None,
    ):
        """Load all SynthDefs from directory.

        Parameters
        ----------
        synthdef_dir : str, optional
            directory with SynthDefs, by default sc3nb default SynthDefs
        completion_msg : bytes, optional
            Message to be executed by the server when loaded, by default None
        server : SCServer, optional
            Server that gets the SynthDefs,
            by default use the SC default server
        """
        if server is None:
            server = sc3nb.SC.get_default().server

        def _load_synthdefs(path):
            cmd_args: List[Union[str, bytes]] = [path.as_posix()]
            if completion_msg is not None:
                cmd_args.append(completion_msg)
            server.msg(
                SynthDefinitionCommand.LOAD_DIR,
                cmd_args,
                await_reply=True,
                bundle=True,
            )

        if synthdef_dir is None:
            ref = libresources.files(sc3nb.resources) / "synthdefs"
            with libresources.as_file(ref) as path:
                _load_synthdefs(path)
        else:
            path = Path(synthdef_dir)
            if path.exists() and path.is_dir():
                _load_synthdefs(path)
            else:
                raise ValueError(f"Provided path {path} does not exist or is not a dir")

    def __init__(self, name: str, definition: str, sc: Optional["SC"] = None) -> None:
        """Create a dynamic synth definition in sc.

        Parameters
        ----------
        name : string
            default name of the synthdef creation.
            The naming convention will be name+int, where int is the amount of
            already created synths of this definition
        definition : string
            Pass the default synthdef definition here. Flexible content
            should be in double brackets ("...{{flexibleContent}}...").
            This flexible content, you can dynamic replace with set_context()
        sc : SC object
            SC instance where the synthdef should be created,
            by default use the default SC instance
        """
        self.sc = sc
        self.definition = definition
        self.name = name
        self.current_def = definition

    def reset(self) -> "SynthDef":
        """Reset the current synthdef configuration to the self.definition value.

        After this you can restart your
        configuration with the same root definition

        Returns
        -------
        object of type SynthDef
            the SynthDef object
        """
        self.current_def = self.definition
        return self

    def set_context(self, searchpattern: str, value) -> "SynthDef":
        """Set context in SynthDef.

        This method will replace a given key (format: "...{{key}}...") in the
        synthdef definition with the given value.

        Parameters
        ----------
        searchpattern : string
            search pattern in the current_def string
        value : string or something with can parsed to string
            Replacement of search pattern

        Returns
        -------
        self : object of type SynthDef
            the SynthDef object
        """
        self.current_def = self.current_def.replace(
            "{{" + searchpattern + "}}", str(value)
        )
        return self

    def set_contexts(self, dictionary: Dict[str, Any]) -> "SynthDef":
        """Set multiple values at onces when you give a dictionary.

        Because dictionaries are unsorted, keep in mind, that
        the order is sometimes ignored in this method.

        Parameters
        ----------
        dictionary : dict
            {searchpattern: replacement}

        Returns
        -------
        self : object of type SynthDef
            the SynthDef object
        """
        for searchpattern, replacement in dictionary.items():
            self.set_context(searchpattern, replacement)
        return self

    def unset_remaining(self) -> "SynthDef":
        """This method will remove all existing placeholders in the current def.

        You can use this at the end of definition
        to make sure, that your definition is clean. Hint: This method will
        not remove pyvars

        Returns
        -------
        self : object of type SynthDef
            the SynthDef object

        """
        self.current_def = re.sub(r"{{[^}]+}}", "", self.current_def)
        return self

    def add(
        self,
        pyvars=None,
        name: Optional[str] = None,
        server: Optional["SCServer"] = None,
    ) -> str:
        """This method will add the current_def to SuperCollider.s

        If a synth with the same definition was already in sc, this method
        will only return the name.

        Parameters
        ----------
        pyvars : dict
            SC pyvars dict, to inject python variables
        name : str, optional
            name which this SynthDef will get
        server : SCServer, optional
            Server where this SynthDef will be send to,
            by default use the SC default server

        Returns
        -------
        str
            Name of the SynthDef
        """
        if name is not None:
            self.name = name

        if self.name in SynthDef.synth_descs:
            del SynthDef.synth_descs[self.name]

        if pyvars is None:
            pyvars = sc3nb.sclang.parse_pyvars(self.current_def)

        # TODO should check if there is context/pyvars that can't be set

        # Create new SynthDef add it to SynthDescLib and get bytes
        if self.sc is None:
            self.sc = sc3nb.SC.get_default()
        synth_def_blob, output = self.sc.lang.cmd(
            f"""
            "sc3nb - Creating SynthDef {self.name}".postln;
            r.tmpSynthDef = SynthDef("{self.name}", {self.current_def});
            SynthDescLib.global.add(r.tmpSynthDef.asSynthDesc);
            r.tmpSynthDef.asBytes();""",
            pyvars=pyvars,
            verbose=False,
            get_result=True,
            get_output=True,
        )
        if synth_def_blob == 0:
            print(output)
            raise RuntimeError(f"Adding SynthDef failed. - {output}")
        else:
            SynthDef.synth_defs[self.name] = synth_def_blob
            if server is not None:
                server.send_synthdef(synth_def_blob)
            else:
                self.sc.server.send_synthdef(synth_def_blob)
            return self.name

    def free(self) -> "SynthDef":
        """Free this SynthDef from the server.

        Returns
        -------
        self : object of type SynthDef
            the SynthDef object
        """
        if self.sc is None:
            self.sc = sc3nb.SC.get_default()
        self.sc.server.msg(SynthDefinitionCommand.FREE, [self.name], bundle=True)
        return self

    def __repr__(self):
        try:
            pyvars = parse_pyvars(self.current_def)
        except NameError:
            current_def = self.current_def
        else:
            current_def = replace_vars(self.current_def, pyvars)
        return f"SynthDef('{self.name}', {current_def})"



# Configuration file for the Sphinx documentation builder.
#
# For a full list of options see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


import json
import os
import shutil
import sys
from pathlib import Path

from git import Repo

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

sys.path.insert(0, os.path.abspath(os.path.join("..", "..")))

# -- Project information -----------------------------------------------------

project = "sc3nb"
copyright = "2021, Thomas Hermann, Dennis Reinsch"
author = "Thomas Hermann, Dennis Reinsch"

# -- General configuration ---------------------------------------------------

on_rtd = os.environ.get("READTHEDOCS") == "True"

repo = Repo(search_parent_directories=True)

git_rev, git_name = repo.commit().name_rev.split()
if "tags/" in git_name:
    # f.e tags/v1.0.1
    git_name = git_name.split("/")[1][1:]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]  # custom css to break long signatures. ref https://github.com/sphinx-doc/sphinx/issues/1514

# -- Options for extensions --------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",  # generate doc
    "sphinx.ext.viewcode",  # generate links to source
    "sphinx.ext.intersphinx",  # allows linking to other projects
    "numpydoc",  # numpy docstring support
    "nbsphinx",  # include notebooks in doc
    "nbsphinx_link",  # support notebooks outside of doc/source via .nblink files
    "sphinx.ext.mathjax",
    "autoapi.extension",  # autoapi from https://github.com/readthedocs/sphinx-autoapi
    "sphinx_rtd_theme",  # read the docs theme
    "myst_parser",  # for supporting markdown
]

suppress_warnings = ["myst.mathjax"]

autosummary_generate = True

numpydoc_validation_checks = {"all", "GL01", "GL02", "GL05"}

intersphinx_mapping = {"python": ("https://docs.python.org/dev", None)}

# -- autoapi -----------------------------------------------------------------

autoapi_type = "python"  # autoapi
autoapi_dirs = ["../../src"]
autoapi_root = "autogen/autoapi"
autoapi_add_toctree_entry = False
autoapi_python_class_content = (
    "both"  # TODO remove this when all classes have the right doc style
)
autoapi_template_dir = "./_templates/autoapi/"


# -- nbsphinx ----------------------------------------------------------------

# nbsphinx_allow_errors = True
nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'png', 'svg', 'pdf'}",
]

# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = (
    r"""
{% if env.metadata[env.docname]['nbsphinx-link-target'] %}
{% set docpath = env.docname |replace("autogen/notebooks", "examples") + ".ipynb" %}
{% else %}
{% set docpath = env.doc2path(env.docname, base='docs/source/') %}
{% endif %}

.. only:: html

    .. role:: raw-html(raw)
        :format: html

    .. nbinfo::
        This page was generated from `{{ docpath }}`__.

    __ https://github.com/interactive-sonification/sc3nb/blob/"""
    + git_rev
    + r"""/{{ docpath }}

.. raw:: latex

    \nbsphinxstartnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{The following section was generated from
    \sphinxcode{\sphinxupquote{\strut {{ docpath | escape_latex }}}} \dotfill}}

"""
)

# This is processed by Jinja2 and inserted after each notebook
nbsphinx_epilog = r"""
{% if env.metadata[env.docname]['nbsphinx-link-target'] %}
{% set docpath = env.docname |replace("autogen/notebooks", "examples") + ".ipynb" %}
{% else %}
{% set docpath = env.doc2path(env.docname, base='docs/source/') %}
{% endif %}

.. raw:: latex

    \nbsphinxstopnotebook{\scriptsize\noindent\strut
    \textcolor{gray}{\dotfill\ \sphinxcode{\sphinxupquote{\strut
    {{ docpath | escape_latex }}}} ends here.}}
"""

# TODO currently readthedocs uses ubuntu18.04, this is old
# it also does not support starting supercollider as it lacks X server support.
# https://docs.readthedocs.io/en/stable/config-file/v2.html#build-apt-packages
if on_rtd:
    nbsphinx_execute = "never"

# -- RTD theme lower left ----------------------------------------------------
# Configuration for the menu on the lower left. Thanks to
# https://stackoverflow.com/questions/49331914/enable-versions-in-sidebar-in-sphinx-read-the-docs-theme
# https://tech.michaelaltfield.net/2020/07/23/sphinx-rtd-github-pages-2/

if not on_rtd:
    html_context = {}
    html_context["display_lower_left"] = True

    current_version = git_name  # tag
    html_context["commit"] = git_rev[:10]

    # tell the theme which version we're currently on ('current_version' affects
    # the lower-left rtd menu and 'version' affects the logo-area version)
    html_context["current_version"] = current_version
    html_context["version"] = current_version

    # POPULATE LINKS TO OTHER VERSIONS
    html_context["versions"] = [
        ("stable", "https://interactive-sonification.github.io/sc3nb/stable/")
    ]
    versions = [branch.name for branch in repo.branches if branch != "gh-pages"]
    for version in versions:
        html_context["versions"].append(
            (
                version,
                "https://interactive-sonification.github.io/sc3nb/" + version + "/",
            )
        )

# -- Custom code -------------------------------------------------------------

# Preparing notebooks


def strip_notebooks(path):
    print("Stripping notebooks")
    extra_keys = "metadata.kernelspec metadata.language_info"
    retval = 0
    for notebook in Path(path).glob("**/*.ipynb"):
        if ".ipynb_checkpoints" not in str(notebook):
            r = os.system(f'nbstripout {str(notebook)} --extra-keys "{extra_keys}"')
            if r == 0:
                print(f"  Stripped {notebook} - {r}")
            else:
                print(f"Error stripping {notebook}")
            retval += r
    return retval


def extract_notebooks_from_doc(path, subdir):
    print("Extracting notebooks from doc source files")
    all_notebooks = []
    for filepath in Path(path).glob("**/*.md"):
        with open(filepath) as file:
            content = file.read()
        notebooks = [
            line.strip().replace(subdir, "")
            for line in content.split("\n")
            if subdir in line
        ]
        if notebooks:
            print(f"  Extracted from {filepath}")
            for nb in notebooks:
                print(f"    {nb}")
        all_notebooks.extend(notebooks)
    return all_notebooks


def generate_notebook_links(
    notebooks_to_link, notebook_dir, doc_dir, link_subdir, media_dir
):
    print("Linking notebooks to doc source")

    nb_dir = Path(notebook_dir).resolve()
    nb_paths = [
        nb_path
        for nb_path in nb_dir.glob("**/*.ipynb")
        if ".ipynb_checkpoints" not in str(nb_path)
    ]
    links_path = Path(doc_dir + link_subdir)
    if links_path.exists():
        shutil.rmtree(links_path, ignore_errors=True)
    linked = []
    for notebook in notebooks_to_link:
        matches = [nb_path for nb_path in nb_paths if notebook in nb_path.as_posix()]
        if len(matches) < 1:
            print(f"> Warning: Could not find {notebook} in {nb_dir}")
        elif len(matches) > 1:
            raise RuntimeError(f"Found {notebook} multiple times {matches}")
        else:
            nb_path = matches[0]
            try:
                nb_path = nb_path.resolve()
                nb_link_path = (
                    (Path(links_path) / nb_path.relative_to(nb_dir))
                    .with_suffix(".nblink")
                    .resolve()
                )
                nb_relative_to_link = Path(
                    os.path.relpath(nb_path, nb_link_path.parent)
                ).as_posix()
                media_dir_relative = Path(
                    os.path.relpath(media_dir, nb_link_path.parent)
                ).as_posix()
                content = {
                    "path": nb_relative_to_link,
                    "extra-media": [media_dir_relative],
                }
                nb_link_path.parent.mkdir(parents=True, exist_ok=True)
                nb_link_path.write_text(json.dumps(content))
            except Exception as excep:
                print(excep)
            else:
                linked.append(nb_path)
                project_dir = Path("../..").resolve()
                print(
                    f"  Linked {nb_path.resolve().relative_to(project_dir)} -> {nb_link_path.resolve().relative_to(project_dir)}"
                )
    for nb in [nb_path.as_posix() for nb_path in nb_paths if nb_path not in linked]:
        print(f"> Warning: Notebook {nb} is not linked")


def prepare_notebooks():
    print("Preparing notebooks...")
    notebooks_dir = "../../examples/"
    notebook_doc_build_subdir = "autogen/notebooks/"
    media_dir = notebooks_dir + "media/"
    source = "./"

    retval = strip_notebooks(notebooks_dir)
    if retval > 0:
        raise RuntimeError("Stripping Notebooks failed.")

    notebooks_to_link = extract_notebooks_from_doc(
        source, subdir=notebook_doc_build_subdir
    )
    generate_notebook_links(
        notebooks_to_link,
        notebooks_dir,
        source,
        notebook_doc_build_subdir,
        media_dir,
    )
    print("Done preparing notebooks.")


prepare_notebooks()



from unittest import TestCase

from sc3nb.sc import startup
from sc3nb.sc_objects.server import ServerOptions


class SCTest(TestCase):
    def test_start_scsynth(self):
        options = ServerOptions(udp_port=57777)
        supercollider = startup(
            start_server=True,
            scsynth_options=options,
            start_sclang=False,
            with_blip=False,
        )

        self.assertIs(supercollider.server.options, options)
        self.assertIsNotNone(supercollider.server)
        with self.assertRaises(RuntimeError):
            supercollider.lang

        supercollider.exit()

    def test_start_sclang(self):
        supercollider = startup(start_server=False, start_sclang=True)

        with self.assertRaises(RuntimeError):
            supercollider.server
        self.assertIsNotNone(supercollider.lang)
        self.assertTrue(supercollider.lang.started)

        supercollider.exit()



import time

from pyamapping import db_to_amp

from sc3nb.sc_objects.node import Synth
from sc3nb.sc_objects.synthdef import SynthDef
from tests.conftest import SCBaseTest


class VolumeTest(SCBaseTest):
    __test__ = True
    start_sclang = True

    def setUp(self) -> None:
        self.assertIsNotNone(VolumeTest.sc.lang)
        VolumeTest.sc.server.unmute()
        self.assertFalse(VolumeTest.sc.server.muted)
        vol_synth = VolumeTest.sc.server._volume._synth
        if vol_synth is not None:
            vol_synth.wait(timeout=1)
        del vol_synth
        self.assertIsNone(VolumeTest.sc.server._volume._synth)
        self.custom_nodeid = 42
        self.all_synth_args = {"freq": 400, "amp": 0.3, "num": 4, "pan": 0, "lg": 0.1}
        self.synth = Synth("s2", nodeid=self.custom_nodeid)

    def tearDown(self) -> None:
        self.synth.free()
        time.sleep(0.1)

    def test_synth_desc(self):
        num_channels = VolumeTest.sc.server.options.num_output_buses
        self.assertIsNotNone(
            SynthDef.get_description(f"sc3nb_volumeAmpControl{num_channels}")
        )

    def test_set_volume(self):
        volume = -10
        VolumeTest.sc.server.volume = volume
        self.assertEqual(volume, VolumeTest.sc.server.volume)
        vol_synth = VolumeTest.sc.server._volume._synth
        self.assertIn(vol_synth, VolumeTest.sc.server.query_tree().children)
        self.assertAlmostEqual(db_to_amp(volume), vol_synth.get("volumeAmp"))
        VolumeTest.sc.server.muted = True
        VolumeTest.sc.server.volume = 0
        self.assertAlmostEqual(0, vol_synth.get("volumeAmp"))
        VolumeTest.sc.server.muted = False
        vol_synth.wait(timeout=0.2)
        self.assertNotIn(vol_synth, VolumeTest.sc.server.query_tree().children)
        del vol_synth
        self.assertFalse(VolumeTest.sc.server.muted)
        self.assertIsNone(VolumeTest.sc.server._volume._synth)



"""Python representation of the scsynth Bus."""

from collections.abc import Iterable
from enum import Enum, unique
from typing import TYPE_CHECKING, Optional, Sequence, Union

import sc3nb
from sc3nb.osc.osc_communication import OSCMessage

if TYPE_CHECKING:
    from sc3nb.sc_objects.server import SCServer


@unique
class ControlBusCommand(str, Enum):
    """OSC Commands for Control Buses"""

    FILL = "/c_fill"
    SET = "/c_set"
    SETN = "/c_setn"
    GET = "/c_get"
    GETN = "/c_getn"


@unique
class BusRate(str, Enum):
    """Calculation rate of Buses"""

    AUDIO = "audio"
    CONTROL = "control"


class Bus:
    """Represenation of Control or Audio Bus(es) on the SuperCollider Server

    If num_channels > 1 this will be represent muliple Buses in a row.

    Parameters
    ----------
    rate : Union[BusRate, str]
        Rate of the Bus, either control or audio
    num_channels : int, optional
        How many channels to allocate, by default 1
    index : int, optional
        Starting Bus index this Bus,
        by default this will be handled by the servers Bus allocator.
    server : SCServer, optional
        Server instance for this Bus,
        by default the default SC server instance.
    """

    def __init__(
        self,
        rate: Union[BusRate, str],
        num_channels: int = 1,
        index: Optional[int] = None,
        server: Optional["SCServer"] = None,
    ) -> None:
        self._server = server or sc3nb.SC.get_default().server
        self._num_channels = num_channels
        self._rate = BusRate(rate)
        if index is None:
            if self._rate is BusRate.AUDIO:
                self._bus_idxs = self._server.audio_bus_ids.allocate(self._num_channels)
            else:
                self._bus_idxs = self._server.control_bus_ids.allocate(
                    self._num_channels
                )
        else:
            self._bus_idxs = list(range(index, index + num_channels))
        if num_channels > 1:
            assert (
                len(self._bus_idxs) == num_channels
            ), "Not enough idxes for number of channels"

    @property
    def rate(self) -> Union[BusRate, str]:
        """The bus calculation rate.

        Returns
        -------
        BusRate
            the rate of this bus
        """
        return self._rate

    @property
    def num_channels(self) -> int:
        """The number of buses.

        Returns
        -------
        int
            number of buses allocated
        """
        return self._num_channels

    @property
    def idxs(self) -> Sequence[int]:
        """The bus index(s).

        Returns
        -------
        int
            first bus index
        """
        return self._bus_idxs

    def is_audio_bus(self) -> bool:
        """Rate check

        Returns
        -------
        bool
            True if this is a audio bus
        """
        return self._rate is BusRate.AUDIO

    def is_control_bus(self) -> bool:
        """Rate check

        Returns
        -------
        bool
            True if this is a control bus
        """
        return self._rate is BusRate.CONTROL

    def set(
        self, *values: Sequence[Union[int, float]], return_msg=False
    ) -> Union["Bus", OSCMessage]:
        """Set ranges of bus values.

        Parameters
        ----------
        values : sequence of int or float
            Values that should be set
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Raises
        ------
        RuntimeError
            If trying to setn an Audio Bus
        """
        if self._rate is BusRate.AUDIO:
            raise RuntimeError("Can't setn Audio Buses")
        if self._num_channels > 1:
            if len(values) != self._num_channels:
                raise ValueError(
                    f"lenght of values must fit num channels ({self._num_channels})"
                )
            msg = OSCMessage(
                ControlBusCommand.SETN, [self._bus_idxs[0], self._num_channels, *values]
            )
        else:
            msg = OSCMessage(ControlBusCommand.SET, [self._bus_idxs[0], *values])
        if return_msg:
            return msg
        else:
            self._server.send(msg, bundle=True)
        return self

    def fill(
        self, value: Union[int, float], return_msg=False
    ) -> Union["Bus", OSCMessage]:
        """Fill bus(es) to one value.

        Parameters
        ----------
        value : Union[int, float]
            value for the buses
        return_msg : bool, optional
            If True return msg else send it directly, by default False

        Raises
        ------
        RuntimeError
            If fill is used on a Audio Bus
        """
        if self._rate is BusRate.AUDIO:
            raise RuntimeError("Can't fill Audio Buses")
        msg = OSCMessage(
            ControlBusCommand.FILL, [self._bus_idxs[0], self._num_channels, value]
        )
        if return_msg:
            return msg
        else:
            self._server.send(msg, bundle=True)
        return self

    def get(self) -> Union[Union[int, float], Sequence[Union[int, float]]]:
        """Get bus value(s).

        Returns
        -------
        bus value or sequence of bus values
            The current value of this bus
            Multiple values if this bus has num_channels > 1

        Raises
        ------
        RuntimeError
            If get is used on an Audio Bus
        """
        if self._rate is BusRate.AUDIO:
            raise RuntimeError("Can't get Audio Buses")
        if self._num_channels > 1:
            msg_params = [self._bus_idxs[0], self._num_channels]
            response = self._server.msg(
                ControlBusCommand.GETN, msg_params, bundle=False
            )
            if isinstance(response, Iterable):
                _, _, *values = response
                return values
        else:
            response = self._server.msg(
                ControlBusCommand.GET, [self._bus_idxs[0]], bundle=False
            )
            if isinstance(response, Iterable):
                _, value = response
                return value

        raise RuntimeError(f"Failed to get right response, got {response}")

    def free(self, clear: bool = True) -> None:
        """Mark this Buses ids as free again

        Parameters
        ----------
        clear : bool, optional
            Reset bus value(s) to 0, by default True
        """
        if self._rate is BusRate.AUDIO:
            self._bus_idxs = self._server.audio_bus_ids.free(self._bus_idxs)
        else:
            if clear:
                self.fill(0)
            self._bus_idxs = self._server.control_bus_ids.free(self._bus_idxs)

    def __del__(self) -> None:
        if self._bus_idxs:
            self.free()

    def __repr__(self) -> str:
        return f"Bus(rate='{self.rate}', ids={self._bus_idxs})"
